"""Contains the main class for a notebook file."""

from __future__ import annotations

import asyncio
import concurrent.futures
import logging
import threading
from collections import defaultdict
from subprocess import DEVNULL  # noqa S404 - Security implications considered
from typing import TYPE_CHECKING, TypedDict

import nbformat
from _frozen_importlib import _DeadlockError
from jupyter_client import AsyncKernelManager, KernelManager
from jupyter_client.kernelspec import NATIVE_KERNEL_NAME, NoSuchKernel
from jupyter_core.paths import jupyter_path

from euporie.core.config import add_setting

if TYPE_CHECKING:
    from typing import Any, Callable, Coroutine, Optional, Union

    from jupyter_client import KernelClient

    from euporie.core.comm.base import KernelTab


log = logging.getLogger(__name__)


class MsgCallbacks(TypedDict, total=False):
    """Typed dictionary for named message callbacks."""

    get_input: "Callable[[str, bool], None]|None"
    set_execution_count: "Callable[[int], None]|None"
    add_output: "Callable[[dict[str, Any]], None]|None"
    clear_output: "Callable[[bool], None]|None"
    done: "Callable[[dict[str, Any]], None]|None"
    set_metadata: "Callable[[tuple[str, ...], Any], None]|None"
    set_status: "Callable[[str], None]|None"
    set_kernel_info: "Callable[[dict[str, Any]], None]|None"
    completeness_status: "Callable[[dict[str, Any]], None]|None"


class Kernel:
    """Runs a notebook kernel and communicates with it asynchronously.

    Has the ability to run itself in it's own thread.
    """

    def _setup_loop(self) -> "None":
        """Set the current loop the the kernel's event loop.

        This method is intended to be run in the kernel thread.
        """
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()

    def __init__(
        self,
        kernel_tab: "KernelTab",
        threaded: "bool" = True,
        allow_stdin: "bool" = False,
        default_callbacks: "Optional[MsgCallbacks]" = None,
    ) -> "None":
        """Called when the :py:class:`Kernel` is initialized.

        Args:
            kernel_tab: The notebook this kernel belongs to
            threaded: If :py:const:`True`, run kernel communication in a separate thread
            allow_stdin: Whether the kernel is allowed to request input
            default_callbacks: The default callbacks to use on recipt of a message

        """
        self.threaded = threaded
        if threaded:
            self.loop = asyncio.new_event_loop()
            self.thread = threading.Thread(target=self._setup_loop)
            self.thread.daemon = True
            self.thread.start()
        else:
            self.loop = asyncio.get_event_loop()

        self.allow_stdin = allow_stdin

        self.kernel_tab = kernel_tab
        self.kc: "Optional[KernelClient]" = None
        self.km = AsyncKernelManager(
            kernel_name=str(kernel_tab.kernel_name),
        )
        self._status = "stopped"
        self.error: "Optional[Exception]" = None

        self.coros: "dict[str, concurrent.futures.Future]" = {}
        self.poll_tasks: "list[asyncio.Task]" = []

        self.default_callbacks = MsgCallbacks(
            {
                "get_input": None,
                "set_execution_count": None,
                "add_output": None,
                "clear_output": None,
                "done": None,
                "set_metadata": None,
                "set_status": None,
            }
        )
        if default_callbacks is not None:
            self.default_callbacks.update(default_callbacks)

        self.msg_id_callbacks: "dict[str, MsgCallbacks]" = defaultdict(
            # Return a copy of the default callbacks
            lambda: MsgCallbacks(dict(self.default_callbacks))  # type: ignore # mypy #8890
        )

        # Set the kernel folder list to prevent the default method from running.
        # This prevents the kernel spec manager from loading IPython, just for the
        # purpose of adding the depreciated :file:`.ipython/kernels` folder to the list
        # of kernel search paths. Without this, having IPython installed causes a
        # import race condition error where IPython was imported in the main thread for
        # displaying LaTeX and in the kernel thread to discover kernel paths.
        # Also this speeds up launch since importing IPython is pretty slow.
        self.km.kernel_spec_manager.kernel_dirs = jupyter_path("kernels")

        self.status_change_event = asyncio.Event()

    def _aodo(
        self,
        coro: "Coroutine",
        wait: "bool" = False,
        callback: "Optional[Callable]" = None,
        timeout: "Optional[Union[int, float]]" = None,
        single: "bool" = False,
    ) -> "Any":
        """Schedules a coroutine in the kernel's event loop.

        Optionally waits for the results (blocking the main thread). Optionally
        schedules a callback to run when the coroutine has completed or timed out.

        Args:
            coro: The coroutine to run
            wait: If :py:const:`True`, block until the kernel has started
            callback: A function to run when the coroutine completes. The result from
                the coroutine will be passed as an argument
            timeout: The number of seconds to allow the coroutine to run if waiting
            single: If :py:const:`True`, any futures for previous instances of the
                coroutine will be cancelled

        Returns:
            The result of the coroutine

        """
        future = asyncio.run_coroutine_threadsafe(coro, self.loop)

        # Cancel previous future instances if required
        if single and self.coros.get(coro.__name__):
            self.coros[coro.__name__].cancel()
        self.coros[coro.__name__] = future

        if wait:
            result = None
            try:
                result = future.result(timeout)
            except concurrent.futures.TimeoutError:
                log.error("Operation '%s' timed out", coro)
                future.cancel()
            finally:
                if callable(callback):
                    callback(result)
            return result
        else:
            if callable(callback):
                future.add_done_callback(
                    lambda f: callback(f.result()) if callback else None
                )

    def _set_living_status(self, alive: "bool") -> "None":
        """Set the life status of the kernel."""
        if not alive:
            self.status = "error"

    @property
    def status(self) -> "str":
        """Retrieve the current kernel status.

        Trigger a kernel life status check when retrieved

        Returns:
            The kernel status

        """
        # Check kernel is alive - use client rather than manager if we have one, as we
        # could be connected to a kernel which we was not started by the manager
        self._aodo(
            self.km.is_alive(),
            timeout=0.2,
            callback=self._set_living_status,
            wait=False,
        )

        return self._status

    @status.setter
    def status(self, value: "str") -> "None":
        """Set the kernel status."""
        self.status_change_event.set()
        self._status = value
        self.status_change_event.clear()

    def wait_for_status(self, status: "str" = "idle") -> "None":
        """Block until the kernel reasches a given status value."""
        if self.status != status:

            async def _wait() -> "None":
                while self.status != status:
                    await asyncio.wait_for(
                        self.status_change_event.wait(), timeout=None
                    )

            self._aodo(_wait(), wait=True)

    @property
    def missing(self) -> "bool":
        """Return True if the requested kernel is not found."""
        try:
            self.km.kernel_spec
        except NoSuchKernel:
            return True
        else:
            return self.km.kernel_spec is None

    @property
    def id(self) -> "Optional[str]":
        """Get the ID of the current kernel."""
        if self.km.has_kernel:
            return self.km.kernel_id
        else:
            return None

    @property
    def specs(self) -> "dict[str, dict]":
        """Returns a list of available kernelspecs."""
        return self.km.kernel_spec_manager.get_all_specs()

    async def stop_(self, cb: "Optional[Callable[[], Any]]" = None) -> "None":
        """Stop the kernel asynchronously."""
        for task in self.poll_tasks:
            task.cancel()
        if self.kc is not None:
            self.kc.stop_channels()
        if self.km.has_kernel:
            await self.km.shutdown_kernel()
        log.debug("Kernel %s shutdown", self.id)

    async def start_(self) -> "None":
        """Start the kernel asynchronously and set its status."""
        if self.km.kernel_name is None:
            self.status = "error"
        log.debug("Starting kernel")
        self.status = "starting"

        # If we are connecting to an existing kernel, create a kernel client using
        # the given connection file
        if self.kernel_tab.app.config.kernel_connection_file:
            connection_file = self.kernel_tab.app.config.kernel_connection_file
            self.km.load_connection_file(connection_file)
            kc = self.km.client_factory(connection_file=connection_file)
            kc.load_connection_file()
            kc.start_channels()
            self.kc = kc

        # Otherwise, start a new kernel using the kernel manager
        else:
            while True:
                try:
                    # TODO - send stdout to log
                    await self.km.start_kernel(stdout=DEVNULL, stderr=DEVNULL)
                except _DeadlockError:
                    # Keep trying if we get an import deadlock
                    continue
                except Exception as e:
                    log.exception("Kernel '%s' could not start", self.km.kernel_name)
                    self.status = "error"
                    self.error = e
                else:
                    log.debug("Started kernel")
                    # Create a client for the newly started kernel
                    if self.km.has_kernel:
                        self.kc = self.km.client()
                break
        try:
            ks = self.km.kernel_spec
        except NoSuchKernel as e:
            self.error = e
            self.status = "error"
        else:
            if ks is not None:
                self.kernel_tab.metadata["kernelspec"] = {
                    "name": self.km.kernel_name,
                    "display_name": ks.display_name,
                    "language": ks.language,
                }

        if self.kc and self.status != "error":
            log.debug("Waiting for kernel to become ready")
            try:
                await self.kc._async_wait_for_ready(timeout=10)
            except RuntimeError as e:
                log.exception("Error connecting to kernel")
                await self.stop_()
                self.error = e
                self.status = "error"
            else:
                log.debug("Kernel %s ready", self.id)
                self.status = "idle"
                self.error = None
                self.poll_tasks = [
                    asyncio.create_task(self.poll("shell")),
                    asyncio.create_task(self.poll("iopub")),
                    asyncio.create_task(self.poll("stdin")),
                ]

    def start(
        self, cb: "Optional[Callable]" = None, wait: "bool" = False, timeout: "int" = 10
    ) -> "None":
        """Starts the kernel.

        Args:
            cb: An optional callback to run after the kernel has started
            wait: If :py:const:`True`, block until the kernel has started
            timeout: How long to wait until failure is assumed

        """
        # Attempt to import ipykernel if it is installed
        # ipykernel is imported by jupyter_client, but since starting the kernel runs
        # in another thread, we do the import here first to prevent import deadlocks,
        # which sometimes occur as we import ipython elsewhere in the main thread
        if self.kernel_tab.kernel_name == NATIVE_KERNEL_NAME:
            try:
                from ipykernel import kernelspec  # noqa F401

                log.debug("Imported `ipykernel` to prevent import deadlock")
            except ImportError:
                pass

        # Start the kernel
        self._aodo(
            self.start_(),
            timeout=timeout,
            wait=wait,
            callback=cb,
        )

    async def poll(self, channel: "str") -> "None":
        """Polls for messages on a channel, and signal when they arrive.

        Args:
            channel: The name of the channel to get messages from

        """
        msg_getter_coro = getattr(self.kc, f"get_{channel}_msg")
        log.debug("Waiting for %s messages", channel)
        while True:
            # log.debug("Waiting for next %s message", channel)
            rsp = await msg_getter_coro()
            # Run msg type handler
            msg_type = rsp.get("header", {}).get("msg_type")
            if callable(handler := getattr(self, f"on_{channel}_{msg_type}", None)):
                handler(rsp)
            else:
                self.on_unhandled(channel, rsp)

    def on_unhandled(self, channel: "str", rsp: "dict[str, Any]") -> "None":
        """Report unhandled messages to the debug log."""
        log.debug(
            "Unhandled %s message:\nparent_id = '%s'\ntype = '%s'\ncontent='%s'",
            channel,
            rsp.get("parent_header", {}).get("msg_id"),
            rsp["header"]["msg_type"],
            rsp.get("content"),
        )

    def on_stdin_input_request(self, rsp: "dict[str, Any]") -> "None":
        """Call ``get_input`` callback for a stdin input request message."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        content = rsp.get("content", {})
        if callable(get_input := self.msg_id_callbacks[msg_id].get("get_input")):
            get_input(
                content.get("prompt", ""),
                content.get("password", False),
            )

    def on_shell_status(self, rsp: "dict[str, Any]") -> "None":
        """Call ``set_execution_count`` callback for a shell status response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        content = rsp.get("content", {})
        status = rsp.get("content", {}).get("status", "")
        if status == "ok":
            if execution_count := content.get("execution_count"):
                if callable(
                    set_execution_count := self.msg_id_callbacks[msg_id].get(
                        "set_execution_count"
                    )
                ):
                    set_execution_count(execution_count)

    def on_shell_execute_reply(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for a shell execute reply response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        content = rsp.get("content", {})

        if self.kernel_tab.app.config.record_cell_timing and callable(
            set_metadata := self.msg_id_callbacks[msg_id]["set_metadata"]
        ):
            set_metadata(
                ("execution", "shell.execute_reply"),
                rsp["header"]["date"].isoformat(),
            )

        if execution_count := content.get("execution_count"):
            if callable(
                set_execution_count := self.msg_id_callbacks[msg_id][
                    "set_execution_count"
                ]
            ):
                set_execution_count(execution_count)

        # Show pager output as a cell execution output
        if payloads := content.get("payload", []):
            if callable(add_output := self.msg_id_callbacks[msg_id]["add_output"]):
                for payload in payloads:
                    if data := payload.get("data", {}):
                        add_output(
                            nbformat.v4.new_output(
                                "execute_result",
                                data=data,
                            )
                        )

        if content.get("status") == "ok":
            if callable(done := self.msg_id_callbacks[msg_id].get("done")):
                done(content)

    def on_shell_kernel_info_reply(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for a shell kernel info response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(
            set_kernel_info := self.msg_id_callbacks[msg_id].get("set_kernel_info")
        ):
            set_kernel_info(rsp.get("content", {}))

    def on_shell_complete_reply(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for a shell completion reply response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(done := self.msg_id_callbacks[msg_id].get("done")):
            done(rsp.get("content", {}))

    def on_shell_history_reply(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for a shell history reply response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(done := self.msg_id_callbacks[msg_id].get("done")):
            done(rsp.get("content", {}))

    def on_shell_inspect_reply(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for a shell inspection reply response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(done := self.msg_id_callbacks[msg_id].get("done")):
            done(rsp.get("content", {}))

    def on_shell_is_complete_reply(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for a shell completeness reply response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(
            completeness_status := self.msg_id_callbacks[msg_id].get(
                "completeness_status"
            )
        ):
            completeness_status(rsp.get("content", {}))

    def on_iopub_status(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for an iopub status response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        status = rsp.get("content", {}).get("execution_state")

        self.status = status
        if callable(set_status := self.msg_id_callbacks[msg_id].get("set_status")):
            set_status(status)

        if status == "idle":
            if self.kernel_tab.app.config.record_cell_timing and callable(
                set_metadata := self.msg_id_callbacks[msg_id].get("set_metadata")
            ):
                set_metadata(
                    ("execution", "iopub.status.idle"),
                    rsp["header"]["date"].isoformat(),
                )

        elif status == "busy":
            if self.kernel_tab.app.config.record_cell_timing and callable(
                set_metadata := self.msg_id_callbacks[msg_id].get("set_metadata")
            ):
                set_metadata(
                    ("execution", "iopub.status.busy"),
                    rsp["header"]["date"].isoformat(),
                )

    def on_iopub_execute_input(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for an iopub execute input response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if self.kernel_tab.app.config.record_cell_timing and callable(
            set_metadata := self.msg_id_callbacks[msg_id]["set_metadata"]
        ):
            set_metadata(
                ("execution", "iopub", "execute_input"),
                rsp["header"]["date"].isoformat(),
            )

    def on_iopub_display_data(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for an iopub display data response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(add_output := self.msg_id_callbacks[msg_id]["add_output"]):
            add_output(nbformat.v4.output_from_msg(rsp))

    def on_iopub_update_display_data(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for an iopub update display data response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(add_output := self.msg_id_callbacks[msg_id]["add_output"]):
            add_output(nbformat.v4.output_from_msg(rsp))

    def on_iopub_execute_result(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for an iopub execute result response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(add_output := self.msg_id_callbacks[msg_id]["add_output"]):
            add_output(nbformat.v4.output_from_msg(rsp))

        if execution_count := rsp.get("content", {}).get("execution_count"):
            if callable(
                set_execution_count := self.msg_id_callbacks[msg_id][
                    "set_execution_count"
                ]
            ):
                set_execution_count(execution_count)

    def on_iopub_error(self, rsp: "dict[str, dict[str, Any]]") -> "None":
        """Call callbacks for an iopub error response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id", "")
        if callable(add_output := self.msg_id_callbacks[msg_id].get("add_output")):
            add_output(nbformat.v4.output_from_msg(rsp))
        if callable(done := self.msg_id_callbacks[msg_id].get("done")):
            done(rsp.get("content", {}))

    def on_iopub_stream(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for an iopub stream response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(add_output := self.msg_id_callbacks[msg_id]["add_output"]):
            add_output(nbformat.v4.output_from_msg(rsp))

    def on_iopub_clear_output(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for an iopub clear output response."""
        # Clear cell output, either now or when we get the next output
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(clear_output := self.msg_id_callbacks[msg_id]["clear_output"]):
            clear_output(rsp.get("content", {}).get("wait", False))

    '''
    def on_iopub_comm_info_reply(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for an comm open response."""
        self.kernel_tab.comm_open(
            content=rsp.get("content", {}), buffers=rsp.get("buffers", [])
        )
    '''

    def on_iopub_comm_open(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for an comm open response."""
        # TODO
        # "If the target_name key is not found on the receiving side, then it should
        # immediately reply with a comm_close message to avoid an inconsistent state."
        #
        self.kernel_tab.comm_open(
            content=rsp.get("content", {}), buffers=rsp.get("buffers", [])
        )

    def on_iopub_comm_msg(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for an iopub comm message response."""
        self.kernel_tab.comm_msg(
            content=rsp.get("content", {}), buffers=rsp.get("buffers", [])
        )

    def on_iopub_comm_close(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for an iopub comm close response."""
        self.kernel_tab.comm_close(
            content=rsp.get("content", {}), buffers=rsp.get("buffers", [])
        )

    def kc_comm(self, comm_id: "str", data: "dict[str, Any]") -> "str":
        """Send a comm message on the shell channel."""
        content = {
            "comm_id": comm_id,
            "data": data,
        }
        if self.kc is not None:
            msg = self.kc.session.msg("comm_msg", content)
            self.kc.shell_channel.send(msg)
            return msg["header"]["msg_id"]
        else:
            raise Exception("Cannot send message when kernel has not started")

    def run(
        self,
        source: "str",
        wait: "bool" = False,
        callback: "Optional[Callable[..., None]]" = None,
        **callbacks: "Callable[..., Any]",
    ) -> "None":
        """Run a cell using the notebook kernel and process the responses."""
        if self.kc is None:
            log.debug("Cannot run cell because kernel has not started")
            # TODO - queue cells for execution
        else:
            self._aodo(
                self.run_(source, **callbacks),
                wait=wait,
                callback=callback,
            )

    async def run_(
        self,
        source: "str",
        get_input: "Optional[Callable[[str, bool], None]]" = None,
        set_execution_count: "Optional[Callable[[int], None]]" = None,
        add_output: "Optional[Callable[[dict[str, Any]], None]]" = None,
        clear_output: "Optional[Callable[[bool], None]]" = None,
        done: "Optional[Callable[[dict[str, Any]], None]]" = None,
        set_metadata: "Optional[Callable[[tuple[str, ...], Any], None]]" = None,
        set_status: "Optional[Callable[[str], None]]" = None,
    ) -> "None":
        """Runs the code cell and and set the response callbacks, optionally waiting."""
        if self.kc is None:
            return
        event = asyncio.Event()
        msg_id = self.kc.execute(
            source,
            store_history=True,
            allow_stdin=self.allow_stdin,
        )

        def wrapped_done(content: "dict[str, Any]") -> "None":
            """Sets the event after the ``done`` callback has completed."""
            # Run the original callback
            if callable(done):
                done(content)
            # Set the event
            event.set()

        callbacks = {
            "get_input": get_input,
            "set_execution_count": set_execution_count,
            "add_output": add_output,
            "clear_output": clear_output,
            "set_metadata": set_metadata,
            "set_status": set_status,
            "done": wrapped_done,
        }
        self.msg_id_callbacks[msg_id].update(
            MsgCallbacks(
                filter(lambda x: x[1] is not None, callbacks.items())
            )  # type: ignore # mypy #8890
        )
        # Wait for "done" callback to be called
        try:
            await asyncio.wait_for(event.wait(), timeout=None)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for kernel info response")

        # Clean up callbacks
        # await asyncio.sleep(0.1)
        # if msg_id in self.msg_id_callbacks:
        # del self.msg_id_callbacks[msg_id]

    def info(
        self,
        set_kernel_info: "Optional[Callable[[dict[str, Any]], None]]" = None,
        set_status: "Optional[Callable[[str], None]]" = None,
    ) -> "None":
        """Request information about the kernel."""
        if self.kc is not None:
            msg_id = self.kc.kernel_info()
            callbacks = {"set_kernel_info": set_kernel_info, "set_status": set_status}
            self.msg_id_callbacks[msg_id].update(
                MsgCallbacks(
                    filter(lambda x: x[1] is not None, callbacks.items())
                )  # type: ignore # mypy #8890
            )

    def comm_info(self, target_name: "Optional[str]" = None) -> "None":
        """Request information about the current comms."""
        if self.kc is not None:
            self.kc.comm_info(target_name=target_name)

    async def complete_(
        self, code: "str", cursor_pos: "int", timeout: "int" = 60
    ) -> "list[dict]":
        """Request code completions from the kernel, asynchronously."""
        results: "list[dict]" = []
        if not self.kc:
            return results

        event = asyncio.Event()

        def process_complete_reply(content: "dict[str, Any]") -> "None":
            """Process response messages on the ``shell`` channel."""
            status = content.get("status", "")
            if status == "ok":
                jupyter_types = content.get("metadata", {}).get(
                    "_jupyter_types_experimental"
                )
                if jupyter_types:
                    for match in jupyter_types:
                        rel_start_position = match.get("start", 0) - cursor_pos
                        completion_type = match.get("type")
                        completion_type = (
                            None if completion_type == "<unknown>" else completion_type
                        )
                        results.append(
                            {
                                "text": match.get("text"),
                                "start_position": rel_start_position,
                                "display_meta": completion_type,
                            }
                        )
                else:
                    rel_start_position = content.get("cursor_start", 0) - cursor_pos
                    for match in content.get("matches", []):
                        results.append(
                            {"text": match, "start_position": rel_start_position}
                        )
                event.set()

        msg_id = self.kc.complete(code, cursor_pos)
        self.msg_id_callbacks[msg_id].update({"done": process_complete_reply})

        try:
            await asyncio.wait_for(event.wait(), timeout)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for kernel info response")

        return results

    def complete(self, code: "str", cursor_pos: "int") -> "list[dict]":
        """Request code completions from the kernel.

        Args:
            code: The code string to retrieve completions for
            cursor_pos: The position of the cursor in the code string

        Returns:
            A list of dictionaries defining completion entries. The dictionaries
            contain ``text`` (the completion text), ``start_position`` (the stating
            position of the completion text), and optionally ``display_meta``
            (a string containing additional data about the completion type)

        """
        return self._aodo(
            self.complete_(code, cursor_pos),
            wait=True,
            single=True,
        )

    async def history_(
        self,
        pattern: "str" = "",
        n: "int" = 1,
        hist_access_type: "str" = "search",
        timeout: "int" = 1,
    ) -> "Optional[list[tuple[int, int, str]]]":
        """Retrieve history from the kernel asynchronously."""
        await asyncio.sleep(0.1)  # Add a tiny timeout so we don't spam the kernel

        results: "list[tuple[int, int, str]]" = []

        if not self.kc:
            return results

        event = asyncio.Event()

        def process_history_reply(content: "dict[str, Any]") -> "None":
            """Process responses on the shell channel."""
            status = content.get("status", "")
            if status == "ok":
                for item in content.get("history", []):
                    results.append(item)
                    event.set()

        msg_id = self.kc.history(
            pattern=pattern, n=n, hist_access_type=hist_access_type, unique=True
        )
        self.msg_id_callbacks[msg_id].update({"done": process_history_reply})

        try:
            await asyncio.wait_for(event.wait(), timeout)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for kernel info response")

        return results

    def history(
        self, pattern: "str" = "", n: "int" = 1, hist_access_type: "str" = "search"
    ) -> "Optional[list[tuple[int, int, str]]]":
        """Retrieve history from the kernel.

        Args:
            pattern: The pattern to search for
            n: the number of history items to return
            hist_access_type: How to access the history ('range', 'tail' or 'search')

        Returns:
            A list of history items, consisting of tuples (session, line_number, input)

        """
        return self._aodo(
            self.history_(pattern, n, hist_access_type),
            wait=True,
            single=True,
        )

    async def inspect_(
        self,
        code: "str",
        cursor_pos: "int",
        detail_level: "int" = 0,
        timeout: "int" = 2,
    ) -> "dict[str, Any]":
        """Retrieve introspection string from the kernel asynchronously."""
        await asyncio.sleep(0.1)  # Add a tiny timeout so we don't spam the kernel

        result: "dict[str, Any]" = {}

        if not self.kc:
            return result

        event = asyncio.Event()

        def process_inspect_reply(content: "dict[str, Any]") -> "None":
            """Process responses on the shell channel."""
            status = content.get("status", "")
            if status == "ok":
                if content.get("found", False):
                    result.update(content)
                    event.set()

        log.debug("Requesting contextual help from the kernel")
        msg_id = self.kc.inspect(code, cursor_pos=cursor_pos, detail_level=detail_level)
        self.msg_id_callbacks[msg_id].update({"done": process_inspect_reply})

        try:
            await asyncio.wait_for(event.wait(), timeout)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for kernel info response")

        return result

    def inspect(
        self,
        code: "str",
        cursor_pos: "int",
        callback: "Optional[Callable[[dict[str, Any]], None]]" = None,
    ) -> "str":
        """Request code inspection from the kernel.

        Args:
            code: The code string to retrieve completions for
            cursor_pos: The position of the cursor in the code string
            callback: A function to run when the inspection result arrives. The result
                is passed as an argument.

        Returns:
            A string containing useful information about the code at the current cursor
            position

        """
        return self._aodo(
            self.inspect_(code, cursor_pos),
            wait=False,
            callback=callback,
            single=True,
        )

    async def is_complete_(
        self,
        code: "str",
        timeout: "Union[int, float]" = 0.1,
    ) -> "dict[str, Any]":
        """Ask the kernel to determine if code is complete asynchronously."""
        result: "dict[str, Any]" = {}

        if not self.kc:
            return result

        event = asyncio.Event()

        def process_is_complete_reply(content: "dict[str, Any]") -> "None":
            """Process responses on the shell channel."""
            result.update(content)
            event.set()

        msg_id = self.kc.is_complete(code)
        self.msg_id_callbacks[msg_id].update(
            {"completeness_status": process_is_complete_reply}
        )

        try:
            await asyncio.wait_for(event.wait(), timeout)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for kernel info response")

        return result

    def is_complete(
        self,
        code: "str",
        timeout: "Union[int, float]" = 0.1,
        wait: "bool" = False,
        callback: "Optional[Callable[[dict[str, Any]], None]]" = None,
    ) -> "dict[str, Any]":
        """Request code completeness status from the kernel.

        Args:
            code: The code string to check the completeness status of
            timeout: How long to wait for a kernel response
            wait: Whether to wait for the response
            callback: A function to run when the inspection result arrives. The result
                is passed as an argument.

        Returns:
            A string describing the completeness status

        """
        return self._aodo(
            self.is_complete_(code, timeout),
            wait=wait,
            callback=callback,
        )

    def interrupt(self) -> "None":
        """Interrupt the kernel.

        This is run in the main thread rather than on the event loop in the kernel's thread,
        because otherwise we would have to wait for currently running tasks on the
        kernel's event loop to finish.
        """
        if self.km.has_kernel:
            log.debug("Interrupting kernel %s", self.id)
            KernelManager.interrupt_kernel(self.km)

    async def restart_(self) -> "None":
        """Restart the kernel asyncchronously."""
        log.debug("Restarting kernel `%s`", self.id)
        await self.km.restart_kernel()
        log.debug("Kernel %s restarted", self.id)

    def restart(self, wait: "bool" = False, cb: "Optional[Callable]" = None) -> "None":
        """Restarts the current kernel."""
        self._aodo(
            self.restart_(),
            wait=wait,
            callback=cb,
        )

    def change(self, name: "str", cb: "Optional[Callable]" = None) -> "None":
        """Change the kernel.

        Args:
            name: The name of the kernel to change to
            cb: Callback to run once restarted

        """
        spec = self.specs.get(name, {}).get("spec", {})

        self.kernel_tab.metadata["kernelspec"] = {
            "name": name,
            "display_name": spec["display_name"],
            "language": spec["language"],
        }
        self.km.kernel_name = name
        if self.km.has_kernel:
            self.restart(cb=cb)
        else:
            self.start(cb=cb)

    def stop(self, cb: "Optional[Callable]" = None, wait: "bool" = False) -> "None":
        """Stops the current kernel.

        Args:
            cb: An optional callback to run when the kernel has stopped.
            wait: If True, wait for the kernel to become idle, otherwise the kernel is
                interrupted before it is stopped

        """
        if self.km.has_kernel is None:
            log.debug("Cannot stop kernel because it is not running")
            if callable(cb):
                cb()
        else:
            log.debug("Stopping kernel %s (wait=%s)", self.id, wait)
            # This helps us leave a little earlier
            if not wait:
                self.interrupt()
            self._aodo(
                self.stop_(),
                callback=cb,
                wait=wait,
            )

    async def shutdown_(self) -> "None":
        """Shut down the kernel and close the event loop if running in a thread."""
        if self.km.has_kernel:
            await self.km.shutdown_kernel(now=True)
        if self.threaded:
            self.loop.stop()
            self.loop.close()
            log.debug("Loop closed")

    def shutdown(self, wait: "bool" = False) -> "None":
        """Shutdown the kernel and close the kernel's thread.

        This is intended to be run when the notebook is closed: the
        :py:class:`~euporie.core.tabs.notebook.Kernel` cannot be restarted after this.

        Args:
            wait: Whether to block until shutdown completes

        """
        self._aodo(
            self.shutdown_(),
            wait=wait,
        )
        if self.threaded:
            self.thread.join(timeout=5)

    # ################################### Settings ####################################

    add_setting(
        name="kernel_connection_file",
        flags=["--kernel-connection-file"],
        type_=str,
        help_="Attempt to connect to an existing kernel using a JSON connection info file",
        default="",
        description="""
            Load connection info from JSON dict. This allows euporie to connect to
            existing kernels.
        """,
    )



"""Contains the main class for a notebook file."""

from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod, abstractproperty
from base64 import standard_b64decode
from functools import partial
from typing import TYPE_CHECKING

import nbformat
from prompt_toolkit.filters import Never, buffer_has_focus

from euporie.core.comm.registry import open_comm
from euporie.core.commands import get_cmd
from euporie.core.config import add_setting
from euporie.core.kernel import MsgCallbacks
from euporie.core.tabs.base import KernelTab
from euporie.core.utils import parse_path
from euporie.core.widgets.cell import Cell, get_cell_id

if TYPE_CHECKING:
    from typing import Any, Callable, Optional

    from prompt_toolkit.filters import Filter
    from prompt_toolkit.layout.containers import AnyContainer
    from upath import UPath

    from euporie.core.app import BaseApp
    from euporie.core.comm.base import Comm
    from euporie.core.kernel import Kernel

log = logging.getLogger(__name__)


class BaseNotebook(KernelTab, metaclass=ABCMeta):
    """The main notebook container class."""

    allow_stdin = False
    edit_mode = False

    def __init__(
        self,
        app: "BaseApp",
        path: "Optional[UPath]" = None,
        kernel: "Optional[Kernel]" = None,
        comms: "Optional[dict[str, Comm]]" = None,
        use_kernel_history: "bool" = False,
        json: "Optional[dict[str, Any]]" = None,
    ):
        """Instantiate a Notebook container, using a notebook at a given path.

        Args:
            path: The file path of the notebook
            app: The euporie application the notebook tab belongs to
            kernel: An existing kernel instance to use
            comms: Existing kernel comm object to use
            use_kernel_history: If True, load history from the kernel
            json: JSON contents of notebook, used if path not given

        """
        self.default_callbacks = MsgCallbacks(
            {
                "get_input": lambda prompt, password: self.cell.get_input(
                    prompt, password
                ),
                "set_execution_count": lambda n: self.cell.set_execution_count(n),
                "add_output": lambda output_json: self.cell.add_output(output_json),
                "clear_output": lambda wait: self.cell.clear_output(wait),
                "set_metadata": lambda path, data: self.cell.set_metadata(path, data),
                "set_status": self.set_status,
                "set_kernel_info": self.set_kernel_info,
            }
        )

        # Load the notebook file
        self.path = parse_path(path)
        log.debug("Loading notebooks %s", self.path)

        self.load(json)

        super().__init__(
            app, path, kernel=kernel, comms=comms, use_kernel_history=use_kernel_history
        )

        self._rendered_cells: "dict[str, Cell]" = {}
        self.load_widgets_from_metadata()
        self.dirty = False
        self.saving = False
        self.multiple_cells_selected: "Filter" = Never()

        self.container = self.load_container()

    # Tab stuff

    def reset(self) -> "None":
        """Reload the notebook file from the disk and re-render."""
        # Restore selection after reset
        if self.path is not None:
            self._rendered_cells = {}
            self.load()
        self.refresh()

    # KernelTab stuff

    @property
    def metadata(self) -> "dict[str, Any]":
        """Return a dictionary to hold notebook / kernel metadata."""
        return self.json.setdefault("metadata", {})

    def kernel_started(self, result: "Optional[dict[str, Any]]" = None) -> "None":
        """Tasks to run when the kernel has started."""
        super().kernel_started(result)

    # Notebook stuff

    def load(self, json: "Optional[dict[str, Any]]" = None) -> "None":
        """Load the notebook file from the file-system."""
        # Open json file, or load from passed json object
        if self.path is not None and self.path.exists():
            with self.path.open() as f:
                self.json = nbformat.read(f, as_version=4)
        else:
            self.json = json or nbformat.v4.new_notebook()
        # Ensure there is always at least one cell
        if not self.json.setdefault("cells", []):
            self.json["cells"] = [nbformat.v4.new_code_cell()]

    def set_status(self, status: "str") -> "None":
        """Called when kernel status changes."""
        self.cell.set_status(status)
        self.app.invalidate()

    @property
    def selected_indices(self) -> "list[int]":
        """Return a list of the currently selected cell indices."""
        return []

    @abstractmethod
    def load_container(self) -> "AnyContainer":
        """Abscract method for loading the notebook's main container."""
        ...

    @abstractproperty
    def cell(self) -> "Cell":
        """Return the current cell."""
        ...

    @property
    def path_name(self) -> "str":
        """Return the path name."""
        if self.path is not None:
            return self.path.name
        else:
            return "(New file)"

    @property
    def title(self) -> "str":
        """Return the tab title."""
        return ("* " if self.dirty else "") + self.path_name

    def lang_file_ext(self) -> "str":
        """Return the file extension for scripts in the notebook's language."""
        return (
            self.json.get("metadata", {})
            .get("language_info", {})
            .get("file_extension", ".py")
        )

    def rendered_cells(self) -> "list[Cell]":
        """Return a list of rendered notebooks' cells."""
        cells = {}
        for i, cell_json in enumerate(self.json.get("cells", [])):
            cell_id = get_cell_id(cell_json)
            if cell_id in self._rendered_cells:
                cells[cell_id] = self._rendered_cells[cell_id]
            else:
                # Pytype doesn't like this..
                cells[cell_id] = Cell(i, cell_json, self)  # type: ignore
            cells[cell_id].index = i
        # These cells will be removed
        for cell in set(self._rendered_cells.values()) - set(cells.values()):
            del cell
        self._rendered_cells = cells
        return list(self._rendered_cells.values())

    def get_cell_by_id(self, cell_id: "str") -> "Optional[Cell]":
        """Returns a reference to the `Cell` container with a given cell id."""
        # Re-render the cells as the one we want might be new
        for cell in self._rendered_cells.values():
            if cell.id == cell_id:
                break
        else:
            return None
        return cell

    def select(
        self,
        cell_index: "int",
        extend: "bool" = False,
        position: "Optional[int]" = None,
        scroll: "bool" = False,
    ) -> "None":
        """Select a cell."""
        pass

    def scroll_to(self, index: "int") -> "None":
        """Scroll to a cell by index."""
        pass

    def refresh(
        self, slice_: "Optional[slice]" = None, scroll: "bool" = False
    ) -> "None":
        """Refresh the notebook."""
        pass

    def close(self, cb: "Optional[Callable]" = None) -> "None":
        """Check if the user want to save an unsaved notebook, then close the file.

        Args:
            cb: A callback to run if after closing the notebook.

        """
        if self.dirty and (unsaved := self.app.dialogs.get("unsaved")):
            unsaved.show(
                tab=self,
                cb=cb,
            )
        else:
            super().close(cb)

    def save(self, path: "UPath" = None, cb: "Optional[Callable]" = None) -> "None":
        """Write the notebook's JSON to the current notebook's file.

        Additionally save the widget state to the notebook metadata.

        Args:
            path: An optional new path at which to save the tab
            cb: A callback to run if after saving the notebook.

        """
        if path is not None:
            self.path = path
        if self.app.config.save_widget_state:
            self.json.setdefault("metadata", {})["widgets"] = {
                "application/vnd.jupyter.widget-state+json": {
                    "version_major": 2,
                    "version_minor": 0,
                    "state": {
                        comm_id: comm._get_embed_state()
                        for comm_id, comm in self.comms.items()
                        if comm.data.get("state", {}).get("__unlinked__") is not True
                    },
                }
            }
        if self.path is None:
            # get_cmd("save-as").run()
            if dialog := self.app.dialogs.get("save-as"):
                dialog.show(tab=self, cb=cb)
        else:
            log.debug("Saving notebook..")
            self.saving = True
            self.app.invalidate()
            try:
                open_file = self.path.open("w")
            except NotImplementedError:
                get_cmd("save-as").run()
            else:
                with open_file as f:
                    nbformat.write(nb=nbformat.from_dict(self.json), fp=f)
                self.dirty = False
                self.saving = False
                self.app.invalidate()
                log.debug("Notebook saved")
            # Run the callback
            if callable(cb):
                cb()

    def run_cell(
        self,
        cell: "Cell",
        wait: "bool" = False,
        callback: "Optional[Callable[..., None]]" = None,
    ) -> "None":
        """Runs a cell.

        Args:
            cell: The rendered cell to run. If ``None``, runs the currently
                selected cell.
            wait: If :py:const:`True`, blocks until cell execution is finished
            callback: Function to run after completion

        """
        cell.remove_outputs()
        self.dirty = True
        # Queue cell if kernel not yet started
        if self.kernel.status == "starting":
            log.debug("Queuing running of cell %s", cell.index)
            self.kernel_queue.append(partial(self.run_cell, cell, wait, callback))
        else:
            self.kernel.run(
                str(cell.json.get("source")),
                wait=wait,
                callback=callback,
                get_input=cell.get_input,
                set_execution_count=cell.set_execution_count,
                add_output=cell.add_output,
                clear_output=cell.clear_output,
                set_metadata=cell.set_metadata,
                set_status=cell.set_status,
                done=cell.ran,
            )

    def load_widgets_from_metadata(self) -> "None":
        """Loads widgets from state saved in notebook metadata."""
        for comm_id, comm_data in (
            self.json.get("metadata", {})
            .get("widgets", {})
            .get("application/vnd.jupyter.widget-state+json", {})
            .get("state", {})
        ).items():
            state = comm_data.get("state", {})
            # Add _model_* keys to the state
            for key, value in comm_data.items():
                if key not in ("state", "buffers"):
                    state[f"_{key}"] = value
            # Flag this widget as not linked to a widget in the kernel, so its state
            # will not be persisted on the next save
            state["__unlinked__"] = True
            # Decode base64 encoded buffers and add buffer paths to data
            buffers = []
            buffer_paths = []
            if "buffers" in state:
                for buffer_data in state.pop("buffers", []):
                    buffer_paths.append(buffer_data["path"])
                    buffers.append(standard_b64decode(buffer_data["data"]))
            # Add this comm to the notebook's current Comm map
            self.comms[comm_id] = open_comm(
                comm_container=self,
                content={
                    "data": {
                        "state": state,
                        "buffer_paths": buffer_paths,
                    },
                    "comm_id": comm_id,
                    "target_name": "jupyter.widget",
                },
                buffers=buffers,
            )

    # ################################### Settings ####################################

    add_setting(
        name="save_widget_state",
        flags=["--save-widget-state"],
        type_=bool,
        help_="Save a notebook's widget state in the notebook metadata",
        default=True,
        description="""
            When set to ``True``, the state of any widgets in the current notebook will
            be saves in the notebook's metadata. This enables widgets to be displayed
            when the notebook is re-opened without having to re-run the notebook.
        """,
    )

    add_setting(
        name="max_notebook_width",
        flags=["--max-notebook-width"],
        type_=int,
        help_="Maximum width of notebooks",
        default=120,
        schema={
            "minimum": 1,
        },
        description="""
            The maximum width at which to display a notebook.
        """,
    )

    add_setting(
        name="expand",
        flags=["--expand"],
        type_=bool,
        help_="Use the full width to display notebooks",
        default=False,
        description="""
            Whether the notebook page should expand to fill the available width
        """,
        cmd_filter=~buffer_has_focus,
    )



"""Contains functions which convert data to formatted ansi text."""

from __future__ import annotations

import logging
from functools import partial
from math import ceil
from typing import TYPE_CHECKING

from euporie.core.app import get_app
from euporie.core.convert.base import register
from euporie.core.convert.formats.common import chafa_convert
from euporie.core.convert.formats.pil import set_background
from euporie.core.convert.utils import call_subproc, commands_exist, have_modules

if TYPE_CHECKING:
    from typing import Any, Optional

    from PIL import Image
    from rich.console import RenderableType
    from upath import UPath

log = logging.getLogger(__name__)


@register(
    from_="html",
    to="ansi",
    filter_=commands_exist("w3m"),
)
def html_to_ansi_w3m(
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Converts HTML text to formatted ANSI using :command:`w3m`."""
    cmd: "list[Any]" = ["w3m", "-T", "text/html"]
    if width is not None:
        cmd += ["-cols", str(width)]
    return call_subproc(data.encode(), cmd).decode()


@register(
    from_="html",
    to="ansi",
    filter_=commands_exist("elinks"),
)
def html_to_ansi_elinks(
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Converts HTML text to formatted ANSI using :command:`elinks`."""
    cmd: "list[Any]" = [
        "elinks",
        "-dump",
        "-dump-color-mode",
        "3",
        "-no-numbering",
        "-force-html",
        "-no-references",
    ]
    if width is not None:
        cmd += ["-dump-width", width]
    return call_subproc(data.encode(), cmd).decode()


@register(
    from_="html",
    to="ansi",
    filter_=commands_exist("lynx"),
)
def html_to_ansi_lynx(
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Converts HTML text to formatted ANSI using :command:`lynx`."""
    cmd: "list[Any]" = ["lynx", "-dump", "-stdin"]
    if width is not None:
        cmd += [f"-width={width}"]
    return call_subproc(data.encode(), cmd).decode()


@register(
    from_="html",
    to="ansi",
    filter_=commands_exist("links"),
)
def html_to_ansi_links(
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Converts HTML text to formatted ANSI using :command:`links`."""
    cmd: "list[Any]" = ["links", "-dump"]
    if width is not None:
        cmd += ["-width", width]
    return call_subproc(data.encode(), cmd, use_tempfile=True).decode()


@register(
    from_="html",
    to="ansi",
    weight=4,
)
def html_to_ansi_py_htmlparser(
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Convert HTML tables to ANSI text using :py:mod:`HTMLParser`."""
    import io
    import re
    from html.parser import HTMLParser

    class HTMLStripper(HTMLParser):
        """Very basic HTML parser which strips style and script tags."""

        def __init__(self):
            super().__init__()
            self.reset()
            self.strict = False
            self.convert_charrefs = True
            self.text = io.StringIO()
            self.skip = False
            self.skip_tags = ("script", "style")

        def handle_starttag(
            self, tag: "str", attrs: "list[tuple[str, Optional[str]]]"
        ) -> "None":
            if tag in self.skip_tags:
                self.skip = True

        def handle_endtag(self, tag: "str") -> "None":
            if tag in self.skip_tags:
                self.skip = False

        def handle_data(self, d: "str") -> "None":
            if not self.skip:
                self.text.write(d)

        def get_data(self) -> "str":
            return self.text.getvalue()

    stripper = HTMLStripper()
    stripper.feed(data)
    output = stripper.get_data()
    # Strip lines
    output = "\n".join([x.strip() for x in output.strip().split("\n")])
    # Remove empty paragraphs
    output = re.sub("\n\n\n+", "\n\n", output)
    return output


@register(
    from_="latex",
    to="ansi",
    filter_=have_modules("flatlatex.latexfuntypes"),
)
def latex_to_ansi_py_flatlatex(
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Convert LaTeX to ANSI using :py:mod:`flatlatex`."""
    import flatlatex

    return flatlatex.converter().convert(data.strip().strip("$").strip())


@register(
    from_="latex",
    to="ansi",
    filter_=have_modules("pylatexenc"),
)
def latex_to_ansi_py_pylatexenc(
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Convert LaTeX to ANSI using :py:mod:`pylatexenc`."""
    from pylatexenc.latex2text import LatexNodes2Text

    return LatexNodes2Text().latex_to_text(data.strip().strip("$").strip())


@register(
    from_="latex",
    to="ansi",
    filter_=have_modules("sympy", "antlr4"),
)
def latex_to_ansi_py_sympy(
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Convert LaTeX to ANSI using :py:mod:`sympy`."""
    from sympy import pretty
    from sympy.parsing.latex import parse_latex
    from sympy.parsing.latex.errors import LaTeXParsingError

    from euporie.core.log import stdout_to_log

    with stdout_to_log(log):
        try:
            parsed = parse_latex(data.strip().strip("$").strip())
        except LaTeXParsingError:
            parsed = data
    return pretty(parsed)


@register(
    from_="pil",
    to="ansi",
    filter_=have_modules("timg"),
)
def pil_to_ansi_py_timg(
    data: "Image",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Convert a PIL image to ANSI text using :py:mod:`timg`."""
    import timg

    w, h = data.size
    if cols is not None:
        data = data.resize((cols, ceil(cols / w * h)))
    bg = bg or get_app().color_palette.bg.base_hex
    if bg:
        data = set_background(data, bg)
    data = set_background(data, bg)
    return timg.Ansi24HblockMethod(data).to_string()


@register(
    from_="pil",
    to="ansi",
    filter_=have_modules("img2unicode"),
)
def pil_to_ansi_py_img2unicode(
    data: "Image",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Convert a PIL image to ANSI text using :py:mod:`img2unicode`."""
    import io

    from img2unicode import FastQuadDualOptimizer, Renderer

    output = io.StringIO()
    Renderer(FastQuadDualOptimizer(), max_w=cols, max_h=rows).render_terminal(
        data, output
    )
    output.seek(0)
    return output.read()


register(
    from_=("png", "jpeg", "svg", "pdf", "gif"),
    to="ansi",
    filter_=commands_exist("chafa"),
)(partial(chafa_convert, "symbols"))


@register(
    from_=("png", "jpeg", "pdf", "gif"),
    to="ansi",
    filter_=commands_exist("timg") & ~have_modules("timg"),
)
def image_to_ansi_timg(
    data: "bytes",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Converts image data to ANSI text using :command:`timg`."""
    cmd: "list[Any]" = ["timg"]
    if cols is not None and rows is not None:
        cmd += [f"-g{cols}x{cols}"]
    cmd += ["--compress", "-pq", "--threads=-1", "-"]
    return call_subproc(data, cmd).decode()


@register(
    from_=("png", "jpeg"),
    to="ansi",
    filter_=commands_exist("catimg"),
)
def image_to_ansi_catimg(
    data: "bytes",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Converts image data to ANSI text using :command:`catimg`."""
    cmd: "list[Any]" = ["catimg"]
    if cols is not None and rows is not None:
        cmd += ["-w", cols * 2]
    cmd += ["-"]
    return call_subproc(data, cmd).decode()


@register(
    from_=("png", "jpeg"),
    to="ansi",
    filter_=commands_exist("icat"),
)
def image_to_ansi_icat(
    data: "bytes",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Converts image data to ANSI text using :command:`icat`."""
    cmd: "list[Any]" = ["icat"]
    if cols is not None and rows is not None:
        cmd += ["-w", cols]
    cmd += ["--mode", "24bit", "-"]
    return call_subproc(data, cmd).decode()


@register(
    from_=("png", "jpeg", "svg", "pdf"),
    to="ansi",
    filter_=commands_exist("tiv"),
)
def image_to_ansi_tiv(
    data: "bytes",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Converts image data to ANSI text using :command:`tiv`."""
    cmd: "list[Any]" = ["tiv"]
    if cols is not None and rows is not None:
        cmd += ["-w", cols, "-h", rows]
    return call_subproc(data, cmd, use_tempfile=True).decode()


@register(
    from_=("png", "jpeg"),
    to="ansi",
    filter_=commands_exist("viu"),
)
def image_to_ansi_viu(
    data: "bytes",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Converts image data to ANSI text using :command:`viu`."""
    cmd: "list[Any]" = ["viu"]
    if cols is not None and rows is not None:
        cmd += ["-w", cols]
    cmd += ["-s", "-"]
    return call_subproc(data, cmd).decode()


@register(
    from_=("png", "jpeg"),
    to="ansi",
    filter_=commands_exist("jp2a"),
)
def image_to_ansi_jp2a(
    data: "bytes",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Converts image data to ANSI text using :command:`jp2a`."""
    cmd: "list[Any]" = ["jp2a", "--color"]
    if cols is not None and rows is not None:
        cmd += [f"--height={rows}"]
    cmd += ["-"]
    return call_subproc(data, cmd).decode()


@register(
    from_="png",
    to="ansi",
    filter_=commands_exist("img2txt"),
)
def png_to_ansi_img2txt(
    data: "bytes",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Converts PNG data to ANSI text using :command:`img2txt`."""
    cmd: "list[Any]" = ["img2txt"]
    if cols is not None and rows is not None:
        cmd += ["-W", cols, "-H", rows]
    return call_subproc(data, cmd, use_tempfile=True).decode()


@register(from_=("png", "jpeg", "svg"), to="ansi", filter_=True, weight=99)
def png_to_ansi_py_placeholder(
    data: "bytes",
    cols: "int" = 7,
    rows: "int" = 3,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Draw placeholder ANSI text."""
    from euporie.core.border import Rounded

    lines = []
    B = Rounded.grid
    lines.append(f"{B.TOP_LEFT}{B.TOP_MID * max(5, (cols - 2))}{B.TOP_RIGHT}")
    lines += [f"{B.MID_LEFT}{B.MID_MID * (cols - 2)}{B.MID_RIGHT}"] * ((rows - 3) // 2)
    lines.append(f"{B.MID_LEFT}{'Image'.center(cols - 2)}{B.MID_RIGHT}")
    lines += [f"{B.MID_LEFT}{B.MID_MID * (cols - 2)}{B.MID_RIGHT}"] * (
        (rows - 3) - (rows - 3) // 2
    )
    lines.append(f"{B.BOTTOM_LEFT}{B.BOTTOM_MID * max(5, (cols - 2))}{B.BOTTOM_RIGHT}")
    return "\n".join(lines)


@register(
    from_="rich",
    to="ansi",
    filter_=have_modules("rich"),
)
def rich_to_ansi_py(
    data: "RenderableType",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Converts rich objects to formatted ANSI text."""
    import rich

    console = rich.get_console()
    options = console.options
    if width is not None:
        options = options.update(max_width=width)
    buffer = console.render(data, options)
    rendered_lines = console._render_buffer(buffer)
    return rendered_lines



"""Contains function which convert data to sixel format."""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from euporie.core.app import get_app
from euporie.core.convert.base import register
from euporie.core.convert.formats.common import chafa_convert, imagemagick_convert
from euporie.core.convert.utils import call_subproc, commands_exist, have_modules

if TYPE_CHECKING:
    from typing import Any, Optional

    from PIL import Image
    from upath import UPath


register(
    from_=("png", "jpeg", "svg", "pdf"),
    to="sixel",
    filter_=commands_exist("chafa"),
)(partial(chafa_convert, "sixel"))


@register(
    from_=("png", "jpeg"),
    to="sixel",
    filter_=commands_exist("img2sixel"),
)
def png_to_sixel_img2sixel(
    data: "bytes",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Converts PNG data to sixels :command:`img2sixel`."""
    bg = bg or get_app().color_palette.bg.base_hex
    cmd: "list[Any]" = ["img2sixel", "-I"]
    if bg:
        cmd += [f"--bgcolor={bg}"]
    if cols is not None:
        px, _ = get_app().term_info.cell_size_px
        cmd += [f"--width={int(cols * px)}"]
    return call_subproc(data, cmd).decode()


register(
    from_=("png", "jpeg", "svg", "pdf"),
    to="sixel",
    filter_=commands_exist("convert", "magick"),
)(partial(imagemagick_convert, "sixel"))


@register(
    from_="pil",
    to="sixel",
    filter_=have_modules("timg"),
)
def pil_to_sixel_py_timg(
    data: "Image",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Convert a pillow image to sixels :py:mod:`timg`."""
    import timg

    return timg.SixelMethod(data).to_string()


@register(
    from_="pil",
    to="sixel",
    filter_=have_modules("teimpy", "numpy"),
)
def pil_to_sixel_py_teimpy(
    data: "Image",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Convert a pillow image to sixels :py:mod:`teimpy`."""
    import numpy as np
    import teimpy

    return teimpy.get_drawer(teimpy.Mode.SIXEL).draw(np.asarray(data))



"""Contains function which convert data to pdf format."""

from __future__ import annotations

from euporie.core.convert.base import register
from euporie.core.convert.formats.common import base64_to_bytes_py

register(
    from_="base64-pdf",
    to="pdf",
)(base64_to_bytes_py)



"""Defines KeyBindings wrapper which keeps track of key binding descriptions."""

from __future__ import annotations

from aenum import extend_enum
from prompt_toolkit.input.ansi_escape_sequences import ANSI_SEQUENCES
from prompt_toolkit.keys import Keys

# Register additional keys
extend_enum(Keys, "ControlEnter", "c-enter")
extend_enum(Keys, "ControlShiftEnter", "c-s-enter")
extend_enum(Keys, "ShiftEnter", "s-enter")

# Assign escape sequences to new keys
ANSI_SEQUENCES["\x1b[27;5;13~"] = Keys.ControlEnter  # type: ignore
ANSI_SEQUENCES["\x1b[13;5u"] = Keys.ControlEnter  # type: ignore

ANSI_SEQUENCES["\x1b[27;2;13~"] = Keys.ShiftEnter  # type: ignore
ANSI_SEQUENCES["\x1b[13;2u"] = Keys.ShiftEnter  # type: ignore

ANSI_SEQUENCES["\x1b[27;6;13~"] = Keys.ControlShiftEnter  # type: ignore
ANSI_SEQUENCES["\x1b[13;6u"] = Keys.ControlShiftEnter  # type: ignore



"""Responsible for loading data from urls."""

import base64
import binascii
import logging
from typing import TYPE_CHECKING
from urllib.parse import urljoin, urlparse, urlunparse
from urllib.request import Request, urlopen

from prompt_toolkit.cache import memoized

if TYPE_CHECKING:
    from typing import Optional, Union

    from upath import UPath

log = logging.getLogger(__name__)


@memoized(maxsize=128)
def load_url(
    url: "Union[UPath, str]", base: "Optional[Union[UPath, str]]" = None
) -> "Optional[bytes]":
    """Loads data from a url."""
    log.debug("Loading data from url `%s`", url)
    data = None

    if base is not None:
        url = urljoin(str(base), str(url))

    parsed_url = urlparse(str(url))

    # If not scheme given, assume it is a local file
    if not parsed_url.scheme:
        parsed_url = parsed_url._replace(scheme="file")
        url = urlunparse(parsed_url)

    elif parsed_url.scheme == "data":
        _mime, _, url_data = parsed_url.path.partition(";")
        data_format, _, encoded_data = url_data.partition(",")
        if data_format == "base64":
            try:
                data = base64.b64decode(encoded_data)
            except binascii.Error:
                data = None
    else:
        request = Request(
            url, headers={"User-Agent": "euporie", "Host": parsed_url.netloc}
        )
        try:
            # The use of 'file:' scheme is intended
            data = urlopen(request, timeout=4).read()  # noqa S310
        except Exception:
            log.debug("Failed to load `%s`", url)

    return data



"""Contains the main class for a notebook file."""

from __future__ import annotations

import logging
from collections import deque
from copy import deepcopy
from functools import partial
from typing import TYPE_CHECKING, cast

import nbformat
from prompt_toolkit.clipboard.base import ClipboardData
from prompt_toolkit.filters import (
    Condition,
    buffer_has_focus,
    has_completions,
    vi_mode,
    vi_navigation_mode,
)
from prompt_toolkit.layout.containers import ConditionalContainer, HSplit, VSplit
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.mouse_events import MouseEventType

from euporie.core.app import get_app
from euporie.core.commands import add_cmd, get_cmd
from euporie.core.config import add_setting
from euporie.core.filters import (
    cursor_on_first_line,
    cursor_on_last_line,
    display_has_focus,
    have_formatter,
    insert_mode,
    kernel_is_python,
    kernel_tab_has_focus,
    multiple_cells_selected,
    replace_mode,
)
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.margins import MarginContainer, ScrollbarMargin
from euporie.core.style import KERNEL_STATUS_REPR
from euporie.core.tabs.base import KernelTab
from euporie.core.tabs.notebook import BaseNotebook
from euporie.core.widgets.cell import Cell
from euporie.core.widgets.decor import Line, Pattern
from euporie.core.widgets.page import ScrollingContainer
from euporie.notebook.filters import (
    cell_has_focus,
    code_cell_selected,
    deleted_cells,
    notebook_has_focus,
)

if TYPE_CHECKING:
    from typing import Any, Deque, MutableSequence, Optional, Sequence

    from prompt_toolkit.formatted_text.base import AnyFormattedText, StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.mouse_events import MouseEvent
    from upath import UPath

    from euporie.core.app import BaseApp
    from euporie.core.comm.base import Comm
    from euporie.core.kernel import Kernel

log = logging.getLogger(__name__)


class Notebook(BaseNotebook):
    """Interactive notebooks.

    A tab which allows running and editing a notebook.
    """

    allow_stdin = True

    def __init__(
        self,
        app: "BaseApp",
        path: "Optional[UPath]" = None,
        kernel: "Optional[Kernel]" = None,
        comms: "Optional[dict[str, Comm]]" = None,
        use_kernel_history: "bool" = True,
        json: "Optional[dict[str, Any]]" = None,
    ) -> "None":
        """Load a editable notebook."""
        super().__init__(
            app=app,
            path=path,
            kernel=kernel,
            comms=comms,
            use_kernel_history=use_kernel_history,
            json=json,
        )

        self.edit_mode = False
        self.in_edit_mode = Condition(self.check_edit_mode)
        self.multiple_cells_selected = multiple_cells_selected
        self.clipboard: "list[Cell]" = []
        self.undo_buffer: "Deque[tuple[int, list[Cell]]]" = deque(maxlen=10)

        if not kernel:
            self.kernel.start(cb=self.kernel_started, wait=False)

    # Tab stuff

    def _statusbar_kernel_handeler(self, event: "MouseEvent") -> "NotImplementedOrNone":
        """Event handler for kernel name field in statusbar."""
        if event.event_type == MouseEventType.MOUSE_UP:
            get_cmd("change-kernel").run()
            return None
        else:
            return NotImplemented

    def statusbar_fields(
        self,
    ) -> "tuple[Sequence[AnyFormattedText], Sequence[AnyFormattedText]]":
        """Generates the formatted text for the statusbar."""
        rendered = self.page.pre_rendered
        return (
            [
                self.mode(),
                f"Cell {self.page.selected_slice.start+1}",
                f"Rendering… ({rendered:.0%})" if rendered < 1 else "",
                "Saving…" if self.saving else "",
            ],
            [
                lambda: cast(
                    "StyleAndTextTuples",
                    [("", self.kernel_display_name, self._statusbar_kernel_handeler)],
                ),
                KERNEL_STATUS_REPR[self.kernel.status] if self.kernel else ".",
            ],
        )

    # Notebook stuff

    @property
    def selected_indices(self) -> "list[int]":
        """Return a list of the currently selected cell indices."""
        return self.page.selected_indices

    def kernel_started(self, result: "Optional[dict[str, Any]]" = None) -> "None":
        """Run when the kernel has started."""
        super().kernel_started(result)
        if not self.kernel_name or self.kernel.missing:
            if not self.kernel_name:
                msg = "No kernel selected"
            else:
                msg = f"Kernel '{self.kernel_display_name}' not installed"
            self.change_kernel(
                msg=msg,
                startup=True,
            )
        elif self.kernel.status == "error":
            self.app.dialogs["error"].show(
                exception=self.kernel.error, when="starting the kernel"
            )
        elif self.kernel.status == "idle":
            if self.app.config.run:
                self.run_all(wait=False)

    def load_container(self) -> "AnyContainer":
        """Load the main notebook container."""
        self.page = ScrollingContainer(
            self.rendered_cells, width=self.app.config.max_notebook_width
        )

        expand = Condition(lambda: self.app.config.expand)

        pattern = Pattern(
            partial(getattr, self.app.config, "background_character"),
            partial(getattr, self.app.config, "background_pattern"),
        )

        return HSplit(
            [
                VSplit(
                    [
                        ConditionalContainer(
                            VSplit(
                                [
                                    pattern,
                                    Line(
                                        char="▋",  # 5/8
                                        width=1,
                                        collapse=True,
                                        style="class:drop-shadow.outer reverse",
                                    ),
                                    Line(
                                        char="▎",  # 2/8
                                        width=1,
                                        collapse=True,
                                        style="class:drop-shadow.inner",
                                    ),
                                ]
                            ),
                            filter=~expand,
                        ),
                        self.page,
                        ConditionalContainer(
                            VSplit(
                                [
                                    Line(
                                        char="▊",  # 6/8
                                        width=1,
                                        collapse=True,
                                        style="class:drop-shadow.inner reverse",
                                    ),
                                    Line(
                                        char="▍",  # 3/8
                                        width=1,
                                        collapse=True,
                                        style="class:drop-shadow.outer",
                                    ),
                                    pattern,
                                ]
                            ),
                            filter=~expand,
                        ),
                        ConditionalContainer(
                            MarginContainer(ScrollbarMargin(), target=self.page),
                            filter=Condition(lambda: self.app.config.show_scroll_bar),
                        ),
                    ],
                    height=Dimension(weight=2),
                ),
            ],
            width=Dimension(weight=1),
            height=Dimension(weight=1),
            key_bindings=load_registered_bindings(
                "euporie.notebook.tabs.notebook.Notebook"
            ),
        )

    @property
    def cell(self) -> "Cell":
        """Returns the currently selected `Cell` in this `Notebook`."""
        cell = self.page.get_child()
        assert isinstance(cell, Cell)
        return cell

    # Editing specific stuff

    @property
    def cells(self) -> "Sequence[Cell]":
        """Returns the currently selected `Cells` in this `Notebook`."""
        return self.page.children[self.page.selected_slice]

    def check_edit_mode(self) -> "bool":
        """Determine if the notebook is (or should be) in edit mode."""
        if self.cell.input_box.has_focus():
            self.edit_mode = True
        return self.edit_mode

    def enter_edit_mode(self) -> "None":
        """Enter cell edit mode."""
        self.edit_mode = True
        self.cell.select()
        self.scroll_to(self.cell.index)

    def exit_edit_mode(self) -> "None":
        """Leave cell edit mode."""
        self.edit_mode = False
        self.cell.select()

    def mode(self) -> "str":
        """Returns a symbol representing the current mode.

        * ``^``: Notebook mode
        * ``>``: Navigation mode
        * ``I``: Insert mode
        * ``v``: Visual mode

        Returns:
            A character representing the current mode
        """
        # TODO - sort this out
        if self.in_edit_mode():
            if insert_mode():
                return "I"
            elif replace_mode():
                return "o"
            else:
                return ">"
        return "^"

    def select(
        self,
        cell_index: "int",
        extend: "bool" = False,
        position: "Optional[int]" = None,
        scroll: "bool" = True,
    ) -> "None":
        """Selects a cell or adds it to the selection.

        Args:
            cell_index: The index of the cell to select
            extend: If true, the selection will be extended to include the cell
            position: An optional cursor position index to apply to the cell input
            scroll: Whether to scroll the page

        """
        # Update the selected slice if we are extending the cell selection
        if extend:
            slice_ = self.page._selected_slice
            stop = -1 if slice_.stop is None else slice_.stop
            step = slice_.step
            if step == -1 and cell_index <= stop:
                stop += 2
                step = 1
            elif step == -1 and cell_index >= stop:
                pass
            elif step in (1, None) and cell_index < stop:
                step = 1
            elif step in (1, None) and cell_index >= stop:
                step = -1
                stop -= 2
            self.page._set_selected_slice(slice(cell_index, stop, step), scroll=scroll)
        # Otherwise set the cell selection to the given cell index
        else:
            self.page._set_selected_slice(
                slice(cell_index, cell_index + 1), scroll=scroll
            )
        # Focus the selected cell - use the current slice in case it did not change
        self.rendered_cells()[self.page.selected_slice.start].focus(
            position, scroll=scroll
        )

    def scroll_to(self, index: "int") -> "None":
        """Scroll to a cell by index."""
        self.page.scroll_to(index)

    def refresh(
        self, slice_: "Optional[slice]" = None, scroll: "bool" = True
    ) -> "None":
        """Refresh the rendered contents of this notebook."""
        if slice_ is None:
            slice_ = self.page._selected_slice
        # This triggers another redraw of the selected cell
        self.page._set_selected_slice(slice_, force=True, scroll=scroll)
        self.page.reset()

    def add_cell_above(self) -> "None":
        """Insert a cell above the current selection."""
        index = self.page.selected_slice.start + 0
        self.add(index)
        self.refresh(slice_=slice(index, index + 1), scroll=False)

    def add_cell_below(self) -> "None":
        """Insert a cell below the current selection."""
        index = self.page.selected_slice.start + 1
        self.add(index)
        self.refresh(slice_=slice(index, index + 1), scroll=True)

    def cut(self, slice_: "Optional[slice]" = None) -> "None":
        """Remove cells from the notebook and them to the `Notebook`'s clipboard."""
        if slice_ is None:
            slice_ = self.page.selected_slice
        self.copy(slice_)
        self.delete(slice_)

    def copy(self, slice_: "Optional[slice]" = None) -> "None":
        """Add a copy of the selected cells to the `Notebook`'s clipboard."""
        if slice_ is None:
            slice_ = self.page.selected_slice
        indices = range(*slice_.indices(len(self.json["cells"])))
        self.clipboard = deepcopy(
            # Sort clipboard contents by index)
            [
                x
                for _, x in sorted(
                    zip(indices, self.json["cells"][slice_]), key=lambda x: x[0]
                )
            ]
        )

    def copy_outputs(self, slice_: "Optional[slice]" = None) -> "None":
        """Copy the outputs of the selected cells."""
        if slice_ is None:
            slice_ = self.page.selected_slice
        output_strings = []

        indices = sorted(range(*slice_.indices(len(self.json["cells"]))))
        rendered_cells = list(self._rendered_cells.values())

        for index in indices:
            cell = rendered_cells[index]
            output_strings.append(cell.output_area.to_plain_text())

        if output_strings:
            self.app.clipboard.set_data(ClipboardData("\n\n".join(output_strings)))

    def paste(self, index: "Optional[int]" = None) -> "None":
        """Append the contents of the `Notebook`'s clipboard below the current cell."""
        if index is None:
            index = self.page.selected_slice.start
        cell_jsons: "MutableSequence" = deepcopy(self.clipboard)
        # Assign a new cell IDs
        for cell_json in cell_jsons:
            cell_json["id"] = nbformat.v4.new_code_cell().get("id")
        self.json["cells"][index + 1 : index + 1] = cell_jsons
        self.dirty = True
        # Only change the selected cell if we actually pasted something
        if cell_jsons:
            self.refresh(slice(index + 1, index + 1 + len(cell_jsons)))

    def add(self, index: "int", source: "str" = "", **kwargs: "Any") -> "None":
        """Creates a new cell at a given index.

        Args:
            index: The position at which to insert a new cell
            source: The contents of the new cell
            kwargs: Additional parameters for the cell

        """
        self.json["cells"].insert(
            index,
            nbformat.v4.new_code_cell(source=source, **kwargs),
        )
        self.dirty = True

    def move(self, n: "int", slice_: "Optional[slice]" = None) -> "None":
        """Move a slice of cells up or down.

        Args:
            slice_: A slice describing the cell indices to move
            n: The amount to move them by

        """
        if slice_ is None:
            slice_ = self.page.selected_slice
        if slice_ is not None:
            indices = range(*slice_.indices(len(self.json["cells"])))
            index = min(indices) + n
            if 0 <= index and index + len(indices) <= len(self.json["cells"]):
                cells = [
                    x
                    for _, x in sorted(
                        zip(indices, self.json["cells"][slice_]), key=lambda x: x[0]
                    )
                ]
                del self.json["cells"][slice_]
                self.json["cells"][index:index] = cells
                self.refresh(
                    slice(
                        slice_.start + n,
                        (-1 if slice_.stop is None else slice_.stop) + n,
                        slice_.step,
                    )
                )
            self.dirty = True

    def delete(self, slice_: "Optional[slice]" = None) -> "None":
        """Delete a cell from the notebook."""
        if slice_ is None:
            slice_ = self.page.selected_slice
        if slice_ is not None:
            indices = range(*slice_.indices(len(self.json["cells"])))
            index = min(indices)
            cell_jsons = [
                x[1] for x in sorted(zip(indices, self.json["cells"][slice_]))
            ]
            self.undo_buffer.append((index, cell_jsons))
            del self.json["cells"][slice_]
            # Ensure there is always one cell
            if len(self.json["cells"]) == 0:
                self.add(1)
            self.dirty = True
            self.refresh(slice(slice_.start, slice_.start + 1), scroll=True)

    def undelete(self) -> "None":
        """Inserts the last deleted cell(s) back into the notebook."""
        if self.undo_buffer:
            index, cells = self.undo_buffer.pop()
            self.json["cells"][index:index] = cells
            self.refresh(slice(index, index + len(cells)))

    def merge(self, slice_: "Optional[slice]" = None) -> "None":
        """Merge two or more cells."""
        if slice_ is None:
            slice_ = self.page.selected_slice
        if slice_ is not None:
            indices = sorted(range(*slice_.indices(len(self.json["cells"]))))
            if len(indices) >= 2:
                # Create a new cell
                new_cell_json = nbformat.v4.new_code_cell()
                # Set the type for that for the focused cell
                cell_type = self.json["cells"][slice_.start].get("cell_type", "code")
                new_cell_json["cell_type"] = cell_type
                # Create and set the combined cell source
                sources = []
                for i in indices:
                    cell_json = self.json["cells"][i]
                    source = cell_json.get("source", "")
                    # Comment markdown cell contents if merging into code cell
                    if cell_type == "code" and cell_json.get("cell_type") == "markdown":
                        source = "\n".join([f"# {line}" for line in source.split("\n")])
                    sources.append(source)
                new_cell_json["source"] = "\n\n".join(sources)
                # Insert the new cell
                new_index = max(indices) + 1
                self.json["cells"].insert(new_index, new_cell_json)
                # Delete the selected slice
                self.delete(slice_)
                self.dirty = True

    def split_cell(self) -> "None":
        """Splits a cell into two at the cursor position."""
        # Get the current cell
        cell = self.cell
        # Get the cursor position in the cell
        cursor_position = cell.input_box.buffer.cursor_position
        # Reset the cursor position to avoid an error after splitting
        cell.input_box.buffer.cursor_position = 0
        # Get the cell contents
        source = cell.json.get("source", "")
        # Create a new cell
        new_cell_json = nbformat.v4.new_code_cell()
        # Split the cell contents at the cursor position
        cell.input = source[cursor_position:]
        new_cell_json["source"] = source[:cursor_position]
        # Copy the cell type
        new_cell_json["cell_type"] = cell.json["cell_type"]
        # Add the new cell to the notebook
        self.json["cells"].insert(cell.index, new_cell_json)
        # Refresh the notebook display
        self.refresh()

    def reformat(self) -> "None":
        """Reformat all code cells in the notebooks."""
        for cell in self.rendered_cells():
            if cell.cell_type == "code":
                cell.reformat()

    def run_selected_cells(
        self,
        advance: "bool" = False,
        insert: "bool" = False,
    ) -> "None":
        """Runs the currently selected cells.

        Args:
            advance: If True, move to next cell. If True and at the last cell, create a
                new cell at the end of the notebook.
            insert: If True, add a new empty cell below the current cell and select it.

        """
        self.exit_edit_mode()
        n_cells = len(self.json["cells"])
        selected_indices = self.page.selected_indices
        cells = {cell.index: cell for cell in self._rendered_cells.values()}

        # Run the cells
        for i in sorted(selected_indices):
            cells[i].run_or_render()
        # Insert a cell if we are at the last cell
        index = max(selected_indices)
        if insert or (advance and max(selected_indices) == (n_cells) - 1):
            self.add(index + 1)
            self.refresh(slice_=slice(index + 1, index + 2), scroll=True)
        elif advance:
            self.select(index + 1)

    def run_all(self, wait: "bool" = False) -> "None":
        """Run all cells."""
        if self.kernel:
            log.debug("Running all cells (wait=%s)", wait)
            for cell in self.rendered_cells():
                if cell.json.get("cell_type") == "code":
                    log.debug("Running cell %s", cell.id)
                    cell.run_or_render(wait=wait)
            log.debug("All cells run")

    # ################################### Settings ####################################

    add_setting(
        name="show_scroll_bar",
        title="scroll bar",
        flags=["--show-scroll-bar"],
        type_=bool,
        help_="Show the scroll bar",
        default=True,
        description="""
            Whether the scroll bar should be shown on the right of the screen.
        """,
    )

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus,
    )
    def _save_notebook() -> "None":
        """Save the current notebook."""
        tab = get_app().tab
        if isinstance(tab, Notebook):
            tab.save()

    @staticmethod
    @add_cmd(
        filter=cell_has_focus & ~buffer_has_focus,
    )
    def _enter_cell_edit_mode() -> "None":
        """Enter cell edit mode."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.enter_edit_mode()

    @staticmethod
    @add_cmd(
        filter=cell_has_focus
        & buffer_has_focus
        & (~vi_mode | (vi_mode & vi_navigation_mode)),
    )
    def _exit_edit_mode() -> "None":
        """Exit cell edit mode."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.exit_edit_mode()

    @staticmethod
    @add_cmd(
        filter=cell_has_focus,
    )
    def _run_selected_cells() -> "None":
        """Run or render the current cells."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.run_selected_cells()

    @staticmethod
    @add_cmd(
        title="Run selected cells and select next cell",
        filter=cell_has_focus,
    )
    def _run_and_select_next() -> "None":
        """Run or render the current cells and select the next cell."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.run_selected_cells(advance=True)

    @staticmethod
    @add_cmd(
        filter=cell_has_focus,
    )
    def _run_cell_and_insert_below() -> "None":
        """Run or render the current cells and insert a new cell below."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.run_selected_cells(insert=True)

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus,
    )
    def _run_all_cells() -> "None":
        """Run or render all the cells in the current notebook."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.run_all()

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _add_cell_above() -> "None":
        """Add a new cell above the current."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.add_cell_above()

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _add_cell_below() -> "None":
        """Add a new cell below the current."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.add_cell_below()

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _delete_cells() -> "None":
        """Delete the current cells."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.delete()

    @staticmethod
    @add_cmd(
        menu_title="Undo delete cell",
        filter=notebook_has_focus
        & ~buffer_has_focus
        & ~display_has_focus
        & deleted_cells,
    )
    def _undelete_cells() -> "None":
        """Undelete the last deleted cells."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.undelete()

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _cut_cells() -> "None":
        """Cut the current cells."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.cut()

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _copy_cells() -> "None":
        """Copy the current cells."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.copy()

    @staticmethod
    @add_cmd(
        menu_title="Copy cell outputs",
        filter=cell_has_focus & ~buffer_has_focus,
    )
    def _copy_outputs() -> "None":
        """Copy the cell's output to the clipboard."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.copy_outputs()

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _paste_cells() -> "None":
        """Paste the previously copied cells."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.paste()

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus
        & ~buffer_has_focus
        & ~display_has_focus
        & multiple_cells_selected,
    )
    def _merge_cells() -> "None":
        """Merge the selected cells."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.merge()

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    @add_cmd(
        filter=notebook_has_focus,
    )
    def _scroll_up() -> "NotImplementedOrNone":
        """Scroll the page up a line."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            return nb.page.scroll(1)
        return None

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    @add_cmd(
        filter=notebook_has_focus,
    )
    def _scroll_down() -> "NotImplementedOrNone":
        """Scroll the page down a line."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            return nb.page.scroll(-1)
        return None

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _scroll_up_5_lines() -> "None":
        """Scroll the page up 5 lines."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.page.scroll(5)

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _scroll_down_5_lines() -> "None":
        """Scroll the page down 5 lines."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.page.scroll(-5)

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _select_first_cell() -> "None":
        """Select the first cell in the notebook."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.select(0)

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _select_5th_previous_cell() -> "None":
        """Go up 5 cells."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.select(nb.page.selected_slice.start - 5)

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _select_previous_cell() -> "None":
        """Go up one cell."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.select(nb.page.selected_slice.start - 1)

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _select_next_cell() -> "None":
        """Select the next cell."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.select(nb.page.selected_slice.start + 1)

    @staticmethod
    @add_cmd(
        filter=~buffer_has_focus,
    )
    def _select_5th_next_cell() -> "None":
        """Go down 5 cells."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.select(nb.page.selected_slice.start + 5)

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _select_last_cell() -> "None":
        """Select the last cell in the notebook."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.select(len(nb.page.children))

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _select_all_cells() -> "None":
        """Select all cells in the notebook."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.page.selected_slice = slice(
                0,
                len(nb.page.children) + 1,
            )

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _extend_cell_selection_to_top() -> "None":
        """Extend the cell selection to the top of the notebook."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.select(0, extend=True)

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _extend_cell_selection_up() -> "None":
        """Extend the cell selection up a cell."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.select(nb.page._selected_slice.start - 1, extend=True)

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _extend_cell_selection_down() -> "None":
        """Extend the cell selection down a cell."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.select(nb.page._selected_slice.start + 1, extend=True)

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _extend_cell_selection_to_bottom() -> "None":
        """Extend the cell selection to the bottom of the notebook."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.select(len(nb.json["cells"]) - 1, extend=True)

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _move_cells_up() -> "None":
        """Move selected cells up."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.move(-1)

    @staticmethod
    @add_cmd(
        filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _move_cells_down() -> "None":
        """Move selected cells down."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.move(1)

    @staticmethod
    @add_cmd(
        filter=cell_has_focus & ~buffer_has_focus,
    )
    def _cells_to_markdown() -> "None":
        """Change selected cells to markdown cells."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            for cell in nb.cells:
                cell.set_cell_type("markdown", clear=True)

    @staticmethod
    @add_cmd(
        filter=cell_has_focus & ~buffer_has_focus,
    )
    def _cells_to_code() -> "None":
        """Change selected cells to code cells."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            for cell in nb.cells:
                cell.set_cell_type("code", clear=False)

    @staticmethod
    @add_cmd(
        filter=cell_has_focus & ~buffer_has_focus,
    )
    def _cells_to_raw() -> "None":
        """Change selected cells to raw cells."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            for cell in nb.cells:
                cell.set_cell_type("raw", clear=True)

    @staticmethod
    @add_cmd(
        filter=cell_has_focus & ~buffer_has_focus,
        title="Expand cell inputs",
    )
    def _show_cell_inputs() -> "None":
        """Expand the selected cells' inputs."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            for cell in nb.cells:
                cell.show_input()

    @staticmethod
    @add_cmd(
        filter=cell_has_focus & ~buffer_has_focus,
        title="Collapse cell inputs",
    )
    def _hide_cell_inputs() -> "None":
        """Collapse the selected cells' inputs."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            for cell in nb.cells:
                cell.hide_input()

    @staticmethod
    @add_cmd(
        filter=cell_has_focus & ~buffer_has_focus,
    )
    def _toggle_cell_inputs() -> "None":
        """Toggle the visibility of the selected cells' inputs."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            for cell in nb.cells:
                cell.toggle_input()

    @staticmethod
    @add_cmd(
        filter=cell_has_focus & ~buffer_has_focus,
        title="Expand cell outputs",
    )
    def _show_cell_outputs() -> "None":
        """Expand the selected cells' outputs."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            for cell in nb.cells:
                cell.show_output()

    @staticmethod
    @add_cmd(
        filter=cell_has_focus & ~buffer_has_focus,
        title="Collapse cell outputs",
    )
    def _hide_cell_outputs() -> "None":
        """Collapse the selected cells' outputs."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            for cell in nb.cells:
                cell.hide_output()

    @staticmethod
    @add_cmd(
        filter=cell_has_focus & ~buffer_has_focus,
    )
    def _toggle_cell_outputs() -> "None":
        """Toggle the visibility of the selected cells' outputs."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            for cell in nb.cells:
                cell.toggle_output()

    @staticmethod
    @add_cmd(
        title="Reformat cells",
        filter=have_formatter
        & code_cell_selected
        & kernel_is_python
        & cell_has_focus
        & ~buffer_has_focus,
    )
    def _reformat_cells() -> "None":
        """Format the selected code cells."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            for cell in nb.cells:
                if cell.cell_type == "code":
                    cell.reformat()

    @staticmethod
    @add_cmd(
        filter=have_formatter
        & kernel_is_python
        & notebook_has_focus
        & ~buffer_has_focus,
    )
    def _reformat_notebook() -> "None":
        """Automatically reformat all code cells in the notebook."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.reformat()

    @staticmethod
    @add_cmd(
        filter=cell_has_focus & ~buffer_has_focus,
    )
    async def _edit_in_external_editor() -> "None":
        """Edit cell in $EDITOR."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            await nb.cell.edit_in_editor()

    @staticmethod
    @add_cmd(
        filter=cell_has_focus & buffer_has_focus,
    )
    def _split_cell() -> "None":
        """Split the current cell at the cursor position."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.split_cell()

    @staticmethod
    @add_cmd(
        filter=cell_has_focus
        & buffer_has_focus
        & cursor_on_first_line
        & ~has_completions,
    )
    def _edit_previous_cell() -> "None":
        """Move the cursor up to the previous cell."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            new_index = nb.cell.index - 1
            cells = nb.rendered_cells()
            if 0 <= new_index < len(cells):
                cells[new_index].select(position=-1, scroll=True)

    @staticmethod
    @add_cmd(
        filter=cell_has_focus
        & buffer_has_focus
        & cursor_on_last_line
        & ~has_completions,
    )
    def _edit_next_cell() -> "None":
        """Move the cursor down to the next cell."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            new_index = nb.cell.index + 1
            cells = nb.rendered_cells()
            if 0 <= new_index < len(cells):
                cells[new_index].select(position=0, scroll=True)

    @staticmethod
    @add_cmd(
        filter=cell_has_focus & ~buffer_has_focus,
    )
    def _scroll_output_left() -> "None":
        """Scroll the cell output to the left."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.cell.output_area.scroll_left()

    @staticmethod
    @add_cmd(
        filter=cell_has_focus & ~buffer_has_focus,
    )
    def _scroll_output_right() -> "None":
        """Scroll the cell output to the right."""
        nb = get_app().tab
        if isinstance(nb, Notebook):
            nb.cell.output_area.scroll_right()

    @staticmethod
    @add_cmd(
        filter=kernel_tab_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _interrupt_kernel() -> "None":
        """Interrupt the notebook's kernel."""
        if isinstance(kt := get_app().tab, KernelTab):
            kt.interrupt_kernel()

    @staticmethod
    @add_cmd(
        filter=kernel_tab_has_focus & ~buffer_has_focus & ~display_has_focus,
    )
    def _restart_kernel() -> "None":
        """Restart the notebook's kernel."""
        if isinstance(kt := get_app().tab, KernelTab):
            kt.restart_kernel()

    @staticmethod
    @add_cmd(
        filter=~buffer_has_focus,
    )
    def _notebook_toggle_line_numbers() -> "None":
        """Toggles line numbers when a buffer does not have focus."""
        get_cmd("toggle-line-numbers").run()

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.notebook.tabs.notebook.Notebook": {
                "save-notebook": "c-s",
                "enter-cell-edit-mode": "enter",
                "exit-edit-mode": "escape",
                "run-selected-cells": ["c-enter", "c-e"],
                "run-and-select-next": ["s-enter", "c-r"],
                "run-cell-and-insert-below": ("escape", "enter"),
                "add-cell-above": "a",
                "add-cell-below": "b",
                "delete-cells": ("d", "d"),
                "undelete-cells": "z",
                "cut-cells": "x",
                "copy-cells": "c",
                "copy-outputs": ("escape", "c"),
                "paste-cells": "v",
                "interrupt-kernel": ("i", "i"),
                "restart-kernel": ("0", "0"),
                "scroll-up": ["[", "<scroll-up>"],
                "scroll-down": ["]", "<scroll-down>"],
                "scroll-up-5-lines": "{",
                "scroll-down-5-lines": "}",
                "select-first-cell": ["home", "c-up"],
                "select-5th-previous-cell": "pageup",
                "select-previous-cell": ["up", "k"],
                "select-next-cell": ["down", "j"],
                "select-5th-next-cell": "pagedown",
                "select-last-cell": ["end", "c-down"],
                "select-all-cells": "c-a",
                "extend-cell-selection-to-top": "s-home",
                "extend-cell-selection-up": ["s-up", "K"],
                "extend-cell-selection-down": ["s-down", "J"],
                "extend-cell-selection-to-bottom": "s-end",
                "move-cells-up": ("escape", "up"),
                "move-cells-down": ("escape", "down"),
                "cells-to-markdown": "m",
                "cells-to-code": "y",
                "cells-to-raw": "r",
                "reformat-cells": "f",
                "reformat-notebook": "F",
                "edit-in-external-editor": "e",
                "merge-cells": "M",
                "split-cell": "c-\\",
                "edit-previous-cell": "up",
                "edit-next-cell": "down",
                "scroll-output-left": "left",
                "scroll-output-right": "right",
                "toggle-expand": "w",
                "notebook-toggle-line-numbers": "l",
                "reset-tab": "f5",
            }
        }
    )



"""This package defines the euporie application and its components."""

__app_name__ = "euporie"
__version__ = "2.1.5"
__logo__ = "⚈"
__strapline__ = "Jupyter in the terminal"
__author__ = "Josiah Outram Halstead"
__email__ = "josiah@halstead.email"
__copyright__ = f"© 2022, {__author__}"
__license__ = "MIT"

from euporie.core.log import default_logs

default_logs()



"""Miscellaneous utility classes."""

from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING, Sequence, TypeVar, overload

from prompt_toolkit.mouse_events import MouseButton, MouseEventType
from upath import UPath

if TYPE_CHECKING:
    from os import PathLike
    from typing import Callable, Iterable, Optional, Union

    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.mouse_handlers import MouseHandler
    from prompt_toolkit.mouse_events import MouseEvent

T = TypeVar("T")


class ChainedList(Sequence[T]):
    """A list-like class which chains multiple lists."""

    def __init__(self, *lists: "Iterable[T]") -> "None":
        """Create a new instance."""
        self.lists = lists

    @property
    def data(self) -> "list[T]":
        """Return the list data."""
        return list(chain.from_iterable(self.lists))

    @overload
    def __getitem__(self, i: "int") -> "T":
        ...

    @overload
    def __getitem__(self, i: "slice"):
        ...

    def __getitem__(self, i):
        """Get an item from the chained lists."""
        return self.data[i]

    def __len__(self) -> "int":
        """Returns the length of the chained lists."""
        return len(self.data)


def parse_path(path: "Optional[Union[str, PathLike]]") -> "Optional[UPath]":
    """Parse and resolve a path."""
    if path is None:
        return None
    upath = UPath(path)
    try:
        upath = upath.expanduser()
    except NotImplementedError:
        pass
    try:
        upath = upath.resolve()
    except AttributeError:
        pass
    return upath


def on_click(func: "Callable") -> "MouseHandler":
    """Return a mouse handler which call a given function on click."""

    def _mouse_handler(mouse_event: "MouseEvent") -> "NotImplementedOrNone":
        if (
            mouse_event.button == MouseButton.LEFT
            and mouse_event.event_type == MouseEventType.MOUSE_UP
        ):
            return func()
        return NotImplemented

    return _mouse_handler



"""A text base user interface for euporie."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, cast

from prompt_toolkit.application.current import get_app as ptk_get_app
from prompt_toolkit.application.run_in_terminal import in_terminal
from prompt_toolkit.filters.app import (
    has_completions,
    is_done,
    renderer_height_is_known,
)
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    FloatContainer,
    HSplit,
    VSplit,
    Window,
)
from prompt_toolkit.layout.dimension import Dimension

from euporie.console.tabs.console import Console
from euporie.core import __logo__
from euporie.core.app import BaseApp
from euporie.core.commands import add_cmd
from euporie.core.config import add_setting
from euporie.core.filters import buffer_is_code, buffer_is_empty, has_dialog
from euporie.core.key_binding.registry import register_bindings
from euporie.core.widgets.dialog import (
    AboutDialog,
    NoKernelsDialog,
    SaveAsDialog,
    SelectKernelDialog,
    ShortcutsDialog,
)
from euporie.core.widgets.pager import Pager
from euporie.core.widgets.palette import CommandPalette
from euporie.core.widgets.search_bar import SearchBar
from euporie.core.widgets.status_bar import StatusBar

if TYPE_CHECKING:
    from typing import Any

log = logging.getLogger(__name__)


def get_app() -> "ConsoleApp":
    """Get the current application."""
    return cast("ConsoleApp", ptk_get_app())


class ConsoleApp(BaseApp):
    """Console app.

    An interactive console which connects to Jupyter kernels and displays
    rich output in the terminal.

    """

    def __init__(self, **kwargs: "Any") -> "None":
        """Create a new euporie text user interface application instance."""
        # Setup mouse support
        self.need_mouse_support = False
        if self.config.mouse_support is not None:
            self.need_mouse_support = self.config.mouse_support

        # Set default application options
        kwargs.setdefault("extend_renderer_height", True)
        kwargs.setdefault("title", "euporie-console")
        kwargs.setdefault("full_screen", False)
        kwargs.setdefault("leave_graphics", True)

        # Initialize the application
        super().__init__(**kwargs)

        self.search_bar = SearchBar()
        self.bindings_to_load += ["euporie.console.app.ConsoleApp"]

        self.tabs = [Console(self)]
        self.pager = Pager()

        self.config.get_item("mouse_support").event += lambda x: setattr(
            self, "need_mouse_support", x.value
        )

    def _get_reserved_height(self) -> "Dimension":
        if has_dialog():
            return Dimension(min=15)
        elif has_completions():
            return Dimension(min=5)
        else:
            return Dimension(min=1)

    def load_container(self) -> "FloatContainer":
        """Returns a container with all opened tabs."""
        assert self.pager is not None
        assert self.search_bar is not None
        assert self.tab is not None

        self.dialogs["command-palette"] = CommandPalette(self)
        self.dialogs["about"] = AboutDialog(self)
        self.dialogs["save-as"] = SaveAsDialog(self)
        self.dialogs["no-kernels"] = NoKernelsDialog(self)
        self.dialogs["change-kernel"] = SelectKernelDialog(self)
        self.dialogs["shortcuts"] = ShortcutsDialog(self)

        return FloatContainer(
            HSplit(
                [
                    self.tab,
                    ConditionalContainer(
                        HSplit(
                            [
                                # Fill empty space below input
                                Window(
                                    height=self._get_reserved_height,
                                    style="class:default",
                                ),
                                self.pager,
                                self.search_bar,
                                VSplit(
                                    [
                                        Window(
                                            char=f" {__logo__} ",
                                            height=1,
                                            width=3,
                                            style="class:menu,logo",
                                            dont_extend_width=True,
                                        ),
                                        StatusBar(),
                                    ]
                                ),
                            ],
                        ),
                        filter=~self.redrawing & ~is_done & renderer_height_is_known,
                    ),
                ]
            ),
            floats=self.floats,  # type: ignore
        )

    def exit(self, **kwargs: "Any") -> "None":
        """Close all tabs on exit."""
        for tab in self.tabs:
            tab.close()
        super().exit(**kwargs)

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd()
    async def _convert_to_notebook() -> "None":
        """Convert the current console session to a notebook."""
        from euporie.notebook.app import NotebookApp
        from euporie.notebook.tabs.notebook import Notebook

        app = get_app()
        nb_app = NotebookApp()
        for tab in app.tabs:
            if isinstance(tab, Console):
                nb = Notebook(
                    app=nb_app,
                    path=tab.path,
                    kernel=tab.kernel,
                    comms=tab.comms,
                    json=tab.json,
                )
                # Set the history to the console's history
                nb.history = tab.history
                # Add the current input
                nb.add(len(nb.json["cells"]) + 1, source=tab.input_box.buffer.text)
                # Add the new notebook to the notebook app
                nb_app.tabs.append(nb)
                # Tell notebook that the kernel has already started
                nb.kernel_started()

        async with in_terminal():
            await nb_app.run_async()

        app.exit()

    @staticmethod
    @add_cmd()
    def _clear_screen() -> "None":
        """Clears the screen and the previous output."""
        app = get_app()
        tab = app.tab
        app.renderer.clear()
        if isinstance(tab, Console):
            tab.reset()
            app.layout.focus(tab.input_box)

    add_cmd(
        name="end-of-file",
        filter=buffer_is_code & buffer_is_empty,
        description="Signals the end of the input, causing the console to exit.",
    )(BaseApp._quit)

    # ################################### Settings ####################################

    add_setting(
        name="mouse_support",
        flags=["--mouse-support"],
        type_=bool,
        help_="Enable or disable mouse support",
        default=None,
        description="""
            When set to True, mouse support is enabled. When set to False, mouse
            support is disabled.
        """,
    )

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.console.app.ConsoleApp": {
                "clear-screen": "c-l",
                "end-of-file": "c-d",
            }
        }
    )



"""Allows drawing tables as :class:`FormattedText`."""

from __future__ import annotations

from collections import defaultdict
from functools import lru_cache, partial
from itertools import tee, zip_longest
from typing import TYPE_CHECKING, cast

from prompt_toolkit.application.current import get_app_session
from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.formatted_text.base import to_formatted_text
from prompt_toolkit.formatted_text.utils import (
    fragment_list_width,
    split_lines,
    to_plain_text,
)

from euporie.core.border import (
    BorderLineStyle,
    GridChar,
    Invisible,
    LineStyle,
    Padding,
    Thin,
    WeightedBorderLineStyle,
    WeightedInt,
    WeightedLineStyle,
    WeightedPadding,
    grid_char,
)
from euporie.core.formatted_text.utils import (
    FormattedTextAlign,
    align,
    max_line_width,
    wrap,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        Hashable,
        Iterable,
        Iterator,
        Optional,
        Sequence,
        TypeVar,
        Union,
    )

    from prompt_toolkit.formatted_text.base import (
        AnyFormattedText,
        FormattedText,
        StyleAndTextTuples,
    )

    PairT = TypeVar("PairT")


def pairwise(iterable: "Iterable[PairT]") -> "Iterator[tuple[PairT, PairT]]":
    """Returns successiver overlapping pairs from an iterable."""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class Cell:
    """A table cell."""

    def _set_padding(self, padding: "Optional[Union[Padding, int]]") -> "None":
        """Sets the cell's padding."""
        if padding is None:
            padding = Padding(None, None, None, None)
        if isinstance(padding, int):
            padding = Padding(padding, padding, padding, padding)
        if len(padding) == 2:
            padding = Padding(padding[0], padding[1], padding[0], padding[1])
        self._padding = padding or Padding(None, None, None, None)

    def _set_border(
        self, border: "Optional[Union[LineStyle, BorderLineStyle]]"
    ) -> "None":
        """Sets the cell's border."""
        if border is None:
            _border = BorderLineStyle(None, None, None, None)
        elif isinstance(border, LineStyle):
            _border = BorderLineStyle(border, border, border, border)
        elif isinstance(border, tuple) and len(border) == 2:
            _border = BorderLineStyle(border[0], border[1], border[0], border[1])
        else:
            _border = border
        self._border = _border

    def __init__(
        self,
        text: "AnyFormattedText" = "",
        row: "Optional[Row]" = None,
        col: "Optional[Col]" = None,
        colspan: "int" = 1,
        rowspan: "int" = 1,
        align: "Optional[FormattedTextAlign]" = None,
        padding: "Optional[Union[Padding, int]]" = None,
        border: "Optional[Union[LineStyle, BorderLineStyle]]" = None,
        style: "str" = "",
        width: "Optional[int]" = None,
    ):
        """Creates a new table cell.

        Args:
            text: Text or formatted text to display in the cell
            row: The row to which this cell belongs
            col: The column to which this cell belongs
            colspan: The number of columns this cell spans
            rowspan: The number of row this cell spans
            align: How the text in the cell should be aligned
            padding: The padding around the contents of the cell
            border: The type of border line to apply to the cell
            style: The style to apply to the cell's contents
            width: The desired width of the cell

        """
        self._text = text
        self.row = row or DummyRow()
        self.col = col or DummyCol()
        self.colspan = colspan
        self.rowspan = rowspan
        self._align = align
        self._set_padding(padding)
        self._set_border(border)
        self._style = ""
        self.style = style
        self._width = width

        self._weighted_border_cache: "SimpleCache[Hashable, WeightedBorderLineStyle]" = (
            SimpleCache()
        )
        self._weighted_padding_cache: "SimpleCache[Hashable, WeightedPadding]" = (
            SimpleCache()
        )

    @property
    def text(self) -> "FormattedText":
        """The cell's input, converted to :class:`FormattedText`."""
        return to_formatted_text(self._text, style=self.style)

    @text.setter
    def text(self, text: "AnyFormattedText") -> "None":
        """Sets the cell's text."""
        self._text = text

    def lines(self, width: "int") -> "Iterable[StyleAndTextTuples]":
        """Wraps the cell's text to a given width.

        Args:
            width: The width at which to wrap the cell's text.

        Returns:
            A list of lines of formatted text
        """
        padding = self.padding
        return split_lines(
            align(
                self.align,
                wrap(
                    [
                        ("", "\n" * (padding.top or 0)),
                        *self.text,
                        ("", "\n" * (padding.bottom or 0)),
                    ],
                    width=width,
                    strip_trailing_ws=True,
                ),
                width=width,
                style=self.style,
            )
        )

    @property
    def width(self) -> "int":
        """The width of the cell excluding padding."""
        if self.colspan > 1:
            return 0
        return self._width or max_line_width(self.text)  # // self.rowspan

    @width.setter
    def width(self, value: "Optional[int]") -> "None":
        """Set the width of the cell."""
        self._width = value

    @property
    def total_width(self) -> "int":
        """The width of the cell including padding."""
        if self.colspan > 1:
            return 0
        padding = self.padding
        return self.width + (padding.left or 0) + (padding.right or 0)

    @property
    def height(self) -> "int":
        """The height of the cell excluding padding."""
        return len(list(split_lines(self.text)))

    @property
    def align(self) -> "FormattedTextAlign":
        """The cell's alignment."""
        if self._align is not None:
            return self._align
        else:
            if self.row._align is not None:
                return self.row.align
            return self.col.align

    @align.setter
    def align(self, align: "FormattedTextAlign") -> "None":
        """Set the cell's alignment."""
        self._align = align

    @property
    def weighted_padding(self) -> "WeightedPadding":
        """The cell's padding with inheritance weights."""

        def _get_weighted_padding() -> "WeightedPadding":
            weights = []
            for i, x in enumerate(self._padding):
                if x is None:
                    weights.append(
                        WeightedInt(
                            max(
                                self.row.weighted_padding[i],
                                self.col.weighted_padding[i],
                            ).weight,
                            max(
                                self.row.weighted_padding[i],
                                self.col.weighted_padding[i],
                            ).value,
                        )
                    )
                else:
                    weights.append(WeightedInt(2, x))
            return WeightedPadding(*weights)

        return self._weighted_padding_cache.get(
            (
                self._padding,
                self.row.weighted_padding,
                self.col.weighted_padding,
            ),
            _get_weighted_padding,
        )

    @property
    def padding(self) -> "Padding":
        """The cell's padding."""
        return cast("Padding", self.weighted_padding.padding)

    @padding.setter
    def padding(self, padding: "Optional[Union[Padding, int]]") -> "None":
        """Sets the cell's padding."""
        self._set_padding(padding)

    @property
    def weighted_border(self) -> "WeightedBorderLineStyle":
        """The cell's borders with inheritance weights."""

        def _get_weighted_border() -> "WeightedBorderLineStyle":
            values = []
            for i, x in enumerate(self._border):
                if i == 1 and self.colspan > 1:
                    # Set right edge to invisible
                    values.append(WeightedLineStyle(3, Invisible))
                elif i == 2 and self.rowspan > 1:
                    # Set bottom edge to invisible
                    values.append(WeightedLineStyle(3, Invisible))
                elif x:
                    values.append(WeightedLineStyle(2, x))
                else:
                    values.append(
                        WeightedLineStyle(
                            max(
                                self.row.weighted_border[i], self.col.weighted_border[i]
                            ).weight,
                            max(
                                self.row.weighted_border[i], self.col.weighted_border[i]
                            ).value,
                        )
                    )
            return WeightedBorderLineStyle(*values)

        return self._weighted_border_cache.get(
            (
                self._border,
                self.row.weighted_border,
                self.col.weighted_border,
                self.colspan,
                self.rowspan,
            ),
            _get_weighted_border,
        )

    @property
    def border(self) -> "BorderLineStyle":
        """The cell's border."""
        return cast("BorderLineStyle", self.weighted_border.border_line_style)

    @border.setter
    def border(self, border: "Optional[Union[LineStyle, BorderLineStyle]]") -> "None":
        """Sets the cell's border."""
        self._set_border(border)

    @property
    def style(self) -> "str":
        """The cell's style."""
        if self.row is None and self.col is None:
            return self._style
        if self.row is None:
            return f"{self.col.style} {self._style}"
        if self.col is None:
            return f"{self.col.style} {self._style}"
        return f"{self.col.style} {self.row.style} {self._style}"

    @style.setter
    def style(self, style: "str") -> "None":
        """Sets the cell's style."""
        self._style = style

    def __repr__(self) -> "str":
        """Returns a text representation of the cell."""
        cell_text = to_plain_text(self._text)
        if len(cell_text) > 5:
            cell_text = cell_text[:4] + "…"
        return f'Cell("{cell_text}")'


class SpacerCell(Cell):
    """A dummy cell to virtually occupy space when ``colspan`` or ``rowspan`` are used."""

    def __init__(
        self,
        expands: "Cell",
        span_index: "int",
        text: "AnyFormattedText" = "",
        row: "Optional[Row]" = None,
        col: "Optional[Col]" = None,
        colspan: "int" = 1,
        rowspan: "int" = 1,
        align: "Optional[FormattedTextAlign]" = None,
        padding: "Optional[Union[Padding, int]]" = None,
        border: "Optional[Union[LineStyle, BorderLineStyle]]" = None,
        style: "str" = "",
        width: "Optional[int]" = None,
    ) -> "None":
        """Creates a new table cell.

        Args:
            expands: This should be reference to the spanning cell
            span_index: The index of a spacer cell inside a colspan / rowspan
            text: Text or formatted text to display in the cell
            row: The row to which this cell belongs
            col: The column to which this cell belongs
            colspan: The number of columns this cell spans
            rowspan: The number of row this cell spans
            align: How the text in the cell should be aligned
            padding: The padding around the contents of the cell
            border: The type of border line to apply to the cell
            style: The style to apply to the cell's contents
            width: The desired width of the cell

        """
        self.expands = expands
        self.span_index = span_index
        super().__init__(
            text="",
            row=row,
            col=col,
            colspan=1,
            rowspan=1,
            align=align,
            padding=0,
            border=expands._border,
            style=expands.style,
            width=0,
        )

    @property
    def weighted_border(self) -> "WeightedBorderLineStyle":
        """The cell's borders with inheritance weights."""
        values = [
            WeightedLineStyle(
                2
                if x
                else max(
                    self.expands.row.weighted_border[i],
                    self.expands.col.weighted_border[i],
                ).weight,
                x
                or max(
                    self.expands.row.weighted_border[i],
                    self.expands.col.weighted_border[i],
                ).value,
            )
            for i, x in enumerate(self.expands._border)
        ]
        if self.expands.colspan > 1:
            # Set left border to invisible in colspan
            values[3] = WeightedLineStyle(3, Invisible)
        if self.expands.rowspan > 1:
            # Set top border to invisible in rowspan
            values[0] = WeightedLineStyle(3, Invisible)
        return WeightedBorderLineStyle(*values)


class RowCol:
    """Base class for table rows and columns."""

    type_: "str"
    span_type: "str"

    def _set_padding(self, padding: "Optional[Union[Padding, int]]") -> "None":
        """The default padding for cells in the row/column."""
        if padding is None:
            padding = Padding(None, None, None, None)
        if isinstance(padding, int):
            padding = Padding(padding, padding, padding, padding)
        if len(padding) == 2:
            padding = Padding(padding[0], padding[1], padding[0], padding[1])
        self._padding = padding

    def _set_border(
        self, border: "Optional[Union[LineStyle, BorderLineStyle]]"
    ) -> "None":
        """Set the default border style for cells in the row/column."""
        if border is None:
            _border = BorderLineStyle(None, None, None, None)
        elif isinstance(border, LineStyle):
            _border = BorderLineStyle(border, border, border, border)
        elif isinstance(border, tuple) and len(border) == 2:
            _border = BorderLineStyle(border[0], border[1], border[0], border[1])
        else:
            _border = border
        self._border = _border

    def __init__(
        self,
        table: "Optional[Table]" = None,
        cells: "Optional[Sequence[Cell]]" = None,
        align: "Optional[FormattedTextAlign]" = None,
        padding: "Optional[Union[Padding, int]]" = None,
        border: "Optional[Union[LineStyle, BorderLineStyle]]" = None,
        style: "str" = "",
    ):
        """Create a new row/column.

        Args:
            table: The :py:class:`table` that this row/column belongs to
            cells: A list of cells in this row/column
            align: The default alignment for cells in this row/column
            padding: The default padding for cells in this row/column
            border: The default border for cells in this row/column
            style: The default style for cells in this row/column

        """
        self.table = table or DummyTable()
        self._cells = defaultdict(lambda: Cell(), enumerate(cells or []))
        if cells:
            for cell in cells:
                setattr(cell, self.type_, self)

        self._align = align
        self._set_padding(padding)
        self._set_border(border)
        self._style = ""
        self.style = style

        self._weighted_border_cache: "SimpleCache[Hashable, WeightedBorderLineStyle]" = (
            SimpleCache()
        )
        self._weighted_padding_cache: "SimpleCache[Hashable, WeightedPadding]" = (
            SimpleCache()
        )

    @property
    def cells(self) -> "list[Cell]":
        """Lists the cells in the row/column."""
        # return [self._cells[i] for i in range(len(self._cells))]
        cells = []
        for i in range(len(self._cells)):
            cell = self._cells[i]
            cells.append(cell)
        return cells

    def new_cell(self, *args: "Any", **kwargs: "Any") -> "Cell":
        """Create a new cell in this row/column."""
        cell = Cell(*args, **kwargs)
        self.add_cell(cell)
        return cell

    def add_cell(self, cell: "Cell") -> "None":
        """Adds a cell to the row/ column."""
        index = max([-1] + list(self._cells.keys())) + 1
        self._cells[index] = cell

        if self.type_ == "row":
            assert isinstance(self, Row)
            cell.row = cast("Row", self)
            cell.col = self.table._cols[index]

            row_index = self.table.rows.index(self)
            col_index = index

            cell.col._cells[row_index] = cell

            if cell.colspan > 1:
                for i in range(1, cell.colspan):
                    spacer = SpacerCell(
                        expands=cell,
                        span_index=i,
                        row=cell.row,
                        col=self.table._cols[col_index + i],
                    )
                    self._cells[col_index + i] = spacer
                    spacer.col._cells[row_index] = spacer

            if cell.rowspan > 1:
                for i in range(1, cell.rowspan):
                    spacer = SpacerCell(
                        expands=cell,
                        span_index=i,
                        row=self.table._rows[row_index + i],
                        col=cell.col,
                    )
                    cell.col._cells[row_index + i] = spacer
                    spacer.row._cells[col_index] = spacer

        elif self.type_ == "col":
            assert isinstance(self, Col)
            cell.row = self.table._rows[index]
            cell.col = cast("Col", self)

            row_index = index
            col_index = self.table.cols.index(self)

            cell.row._cells[col_index] = cell

            if cell.rowspan > 1:
                for i in range(1, cell.rowspan):
                    spacer = SpacerCell(
                        expands=cell,
                        span_index=i,
                        row=self.table._rows[index + i],
                        col=cell.col,
                    )
                    self._cells[row_index + i] = spacer
                    spacer.row._cells[col_index] = spacer

            if cell.colspan > 1:
                index = max(cell.row._cells.keys())
                for i in range(1, cell.colspan):
                    spacer = SpacerCell(
                        expands=cell,
                        span_index=i,
                        row=cell.row,
                        col=self.table._cols[col_index + i],
                    )
                    cell.row._cells[col_index + i] = spacer
                    spacer.col._cells[row_index] = spacer

    @property
    def widths(self) -> "list[int]":
        """A list of the widths of cell (excluding horizontal padding)."""
        return [cell.width for cell in self.cells]

    @property
    def max_width(self) -> "int":
        """The maximum cell width (excluding horizontal padding)."""
        if self.widths:
            return max(self.widths)
        else:
            return 0

    @property
    def total_widths(self) -> "list[int]":
        """A list of cell widths (including horizontal padding)."""
        return [cell.total_width for cell in self.cells]

    @property
    def max_total_width(self) -> "int":
        """The maximum cell width (including horizontal padding)."""
        if self.widths:
            return max(self.total_widths)
        else:
            return 0

    @property
    def align(self) -> "FormattedTextAlign":
        """The default alignment for cells in the row/column."""
        return self._align or self.table.align

    @align.setter
    def align(self, align: "FormattedTextAlign") -> "None":
        """Set the default alignment for cells in the row/column."""
        self._align = align

    @property
    def weighted_padding(self) -> "WeightedPadding":
        """The default padding for cells in the row/column with inheritance weights."""

        def _get_weighted_padding() -> "WeightedPadding":
            return WeightedPadding(
                *(
                    WeightedInt(
                        1 if x is not None else 0,
                        x if x is not None else self.table.padding[i] or 0,
                    )
                    for i, x in enumerate(self._padding)
                )
            )

        return self._weighted_padding_cache.get(
            (self._padding, self.table.padding), _get_weighted_padding
        )

    @property
    def padding(self) -> "Padding":
        """The default padding for cells in the row/column."""
        return cast("Padding", self.weighted_padding.padding)

    @padding.setter
    def padding(self, padding: "Optional[Union[Padding, int]]") -> "None":
        """The default padding for cells in the row/column."""
        self._set_padding(padding)

    @property
    def weighted_border(self) -> "WeightedBorderLineStyle":
        """The cell's borders with inheritance weights."""

        def _get_weighted_border() -> "WeightedBorderLineStyle":
            return WeightedBorderLineStyle(
                *(
                    WeightedLineStyle(
                        1 if x is not None else 0,
                        x if x is not None else self.table.border[i] or Thin,
                    )
                    for i, x in enumerate(self._border)
                )
            )

        return self._weighted_border_cache.get(
            (self._border, self.table.border), _get_weighted_border
        )

    @property
    def border(self) -> "BorderLineStyle":
        """The default border style for cells in the row/column."""
        return cast("BorderLineStyle", self.weighted_border.border_line_style)

    @border.setter
    def border(self, border: "Optional[Union[LineStyle, BorderLineStyle]]") -> "None":
        self._set_border(border)

    @property
    def style(self) -> "str":
        """Set the default style for cells' contents in the row or column."""
        return f"{self.table.style} {self._style}"

    @style.setter
    def style(self, style: "str") -> "None":
        """The default style for cells' contents in the row or column."""
        self._style = style

    def __repr__(self) -> "str":
        """Returns a textual representation of the row or column."""
        return f"{self.__class__.__name__}({', '.join(map(str, self._cells.values()))})"


class Row(RowCol):
    """A row in a table."""

    type_ = "row"
    span_type = "colspan"


class Col(RowCol):
    """A column in a table."""

    type_ = "col"
    span_type = "rowspan"


class DummyCell(Cell):
    """A dummy cell with not content, padding or borders."""

    def __repr__(self) -> "str":
        return f"{self.__class__.__name__}()"


class Table:
    """A table."""

    def _set_border(
        self, border: "Optional[Union[LineStyle, BorderLineStyle]]"
    ) -> "None":
        """Set the default border style for cells in the table."""
        if border is None:
            border = Thin
        if isinstance(border, LineStyle):
            border = BorderLineStyle(border, border, border, border)
        if len(border) == 2:
            border = BorderLineStyle(border[0], border[1], border[0], border[1])
        # None is not a permitted value here - replace with default
        self._border = BorderLineStyle(
            border[0] or Thin,
            border[1] or Thin,
            border[2] or Thin,
            border[3] or Thin,
        )

    def _set_padding(self, padding: "Optional[Union[Padding, int]]") -> "None":
        """Set the default padding for cells in the table."""
        if padding is None:
            padding = Padding(0, 1, 0, 1)
        if isinstance(padding, int):
            padding = Padding(padding, padding, padding, padding)
        if len(padding) == 2:
            padding = Padding(padding[0], padding[1], padding[0], padding[1])
        # `None` is not permitted for padding here, as there is nothing to inherit from
        self._padding = Padding(
            padding[0] or 0, padding[1] or 0, padding[2] or 0, padding[3] or 0
        )

    def __init__(
        self,
        rows: "Optional[Sequence[Row]]" = None,
        cols: "Optional[Sequence[Col]]" = None,
        width: "Optional[int]" = None,
        expand: "bool" = False,
        align: "FormattedTextAlign" = FormattedTextAlign.LEFT,
        padding: "Optional[Union[Padding, int]]" = None,
        border: "Optional[Union[BorderLineStyle, LineStyle]]" = Thin,
        border_style: "str" = "",
        border_collapse: "bool" = False,
        style: "str" = "",
    ) -> "None":
        """Creates a new table instance.

        Args:
            rows: A list of :class:`Row`s to add to the table
            cols: A list of :class:`Col`s to add to the table. Cells specified in rows
                take priority, so cells in column positions already defined by rows
                will be ignored.
            width: The width of the table
            expand: Whether the table should expand to fill the available space
            align: The default alignment for cells in the table
            padding: The default padding for cells in the table
            border: The default border style for cells in the table
            border_style: The style to apply to the table's borders
            border_collapse: If :const:`True`, if horizontal border in the table are
                :class:`Invisible` for their entire length, no extra line will be drawn
            style: A style to apply to the table's cells' contents

        """
        self._rows = defaultdict(partial(Row, self), enumerate(rows or []))
        if rows:
            for row in rows:
                row.table = self
        self._cols = defaultdict(partial(Col, self), enumerate(cols or []))
        if cols:
            for col in cols:
                col.table = self
        if rows:
            self.sync_rows_to_cols()
        elif cols:
            self.sync_cols_to_rows()

        self.width = width or get_app_session().output.get_size()[1]
        self.align = align
        self.expand = expand

        self._set_padding(padding)
        self._set_border(border)
        self.border_style = f"class:border {border_style}"
        self._border_collapse = border_collapse
        self.style = style

        self._border_collapse_cache: "SimpleCache[Hashable, bool]" = SimpleCache()

    @property
    def border_collapse(self) -> "bool":
        """Turn off border-collapse if any cells have a border."""

        def _get_border_collapse() -> "bool":
            for row in self._rows.values():
                for cell in row._cells.values():
                    if any(x != Invisible for x in cell.border):
                        return False
            return self._border_collapse

        return self._border_collapse_cache.get(tuple(self.rows), _get_border_collapse)

    @property
    def border(self) -> "BorderLineStyle":
        """The default border style for cells in the table."""
        return self._border

    @border.setter
    def border(self, border: "Optional[Union[LineStyle, BorderLineStyle]]") -> "None":
        self._set_border(border)

    @property
    def padding(self) -> "Padding":
        """The default padding for cells in the table."""
        return self._padding

    @padding.setter
    def padding(self, padding: "Optional[Union[Padding, int]]") -> "None":
        self._set_padding(padding)

    @property
    def rows(self) -> "list[Row]":
        """A list of rows in the table."""
        return [self._rows[i] for i in range(len(self._rows))]

    @property
    def cols(self) -> "list[Col]":
        """A list of columns in the table."""
        return [self._cols[i] for i in range(len(self._cols))]

    def sync_rows_to_cols(self) -> "None":
        """Ensure cells in rows are present in the relevant columns."""
        cols = self._cols
        for i, row in self._rows.items():
            for j, cell in row._cells.items():
                cell.col = cols[j]
                cols[j]._cells[i] = cell

    def sync_cols_to_rows(self) -> "None":
        """Ensure cells in columns are present in the relevant rows."""
        rows = self._rows
        for i, col in self._cols.items():
            for j, cell in col._cells.items():
                cell.row = rows[j]
                rows[j]._cells[i] = cell

    def new_row(self, *args: "Any", **kwargs: "Any") -> "Row":
        """Creates a new row in the table."""
        row = Row(*args, **kwargs)
        self.add_row(row)
        return row

    def add_row(self, row: "Row") -> "None":
        """Add a row to the table."""
        row.table = self

        unfilled = [
            i
            for i, row in self._rows.items()
            if all([isinstance(cell, SpacerCell) for cell in row._cells.values()])
        ]

        index = min(
            min(unfilled + [len(self._rows)]),
            max([-1] + list(self._rows)) + 1,
        )

        # Merge existing row with new row
        cells = self._rows[index]._cells
        i = 0
        for cell in row.cells:
            while i in cells:
                i += 1
            cells[i] = cell
        row._cells = cells
        self._rows[index] = row

    def new_col(self, *args: "Any", **kwargs: "Any") -> "Col":
        """Creates a new column in the table."""
        col = Col(*args, **kwargs)
        self.add_col(col)
        return col

    def add_col(self, col: "Col") -> "None":
        """Add a column to the table."""
        col.table = self

        unfilled = [
            i
            for i, col in self._cols.items()
            if all([isinstance(cell, SpacerCell) for cell in col._cells.values()])
        ]

        index = min(
            min(unfilled + [len(self._cols)]),
            max([-1] + list(self._cols)) + 1,
        )

        # Merge existing col with new col
        cells = self._cols[index]._cells
        i = 0
        for cell in col.cells:
            while i in cells:
                i += 1
            cells[i] = cell
        col._cells = cells
        self._cols[index] = col

    def calculate_col_widths(
        self, width: "int", min_col_width: "int" = 10
    ) -> "list[int]":
        """Calculate column widths given the available space.

        Reduce the widest column until we fit in available width, or expand cells to
        to fill the available witdth.

        Args:
            width: The desired width of the table
            min_col_width: The minimum width allowed for a column

        Returns:
            List of new column widths

        """
        # TODO - this function is too slow
        col_widths = [col.max_total_width for col in self.cols]

        def total_width(col_widths: "list[int]") -> "int":
            return sum(col_widths) + len(self.cols) + 1

        if self.expand:
            while total_width(col_widths) < max(
                width, len(col_widths) * min_col_width + 2
            ):
                idxmin = min(enumerate(col_widths), key=lambda x: x[1])[0]
                col_widths[idxmin] += 1
        else:
            while total_width(col_widths) > max(
                width, len(col_widths) * min_col_width + 2
            ):
                idxmax = max(enumerate(col_widths), key=lambda x: x[1])[0]
                col_widths[idxmax] -= 1

        return col_widths

    @staticmethod
    @lru_cache
    def get_node(
        nw_wb: "WeightedBorderLineStyle",
        ne_wb: "WeightedBorderLineStyle",
        se_wb: "WeightedBorderLineStyle",
        sw_wb: "WeightedBorderLineStyle",
    ) -> "str":
        """Calculate which character to use at the intersection of four cells."""
        return grid_char(
            GridChar(
                max(nw_wb.right, ne_wb.left).value,
                max(ne_wb.bottom, se_wb.top).value,
                max(se_wb.left, sw_wb.right).value,
                max(sw_wb.top, nw_wb.bottom).value,
            )
        )

    @staticmethod
    @lru_cache
    def get_horizontal_edge(
        n_wb: "WeightedBorderLineStyle", s_wb: "WeightedBorderLineStyle"
    ) -> "str":
        """Calculate which character to use to divide horizontally adjacent cells."""
        line_style = max(n_wb.bottom, s_wb.top).value
        return grid_char(GridChar(Invisible, line_style, Invisible, line_style))

    @staticmethod
    @lru_cache
    def get_vertical_edge(
        e_wb: "WeightedBorderLineStyle", w_wb: "WeightedBorderLineStyle"
    ) -> "str":
        """Calculate which character to use to divide vertically adjacent cells."""
        line_style = max(e_wb.right, w_wb.left).value
        return grid_char(GridChar(line_style, Invisible, line_style, Invisible))

    def draw_border_row(
        self,
        row_above: "Optional[Row]",
        row_below: "Optional[Row]",
        col_widths: "list[int]",
        border_collapse: "bool",
    ) -> "StyleAndTextTuples":
        """Draws a border line separating two rows in the table."""
        output: "StyleAndTextTuples" = []
        if row_above is not None:
            row_above_cells = row_above.cells
        else:
            assert row_below is not None
            row_above_cells = [DummyCell()] * len(row_below.cells)
        if row_below is not None:
            row_below_cells = row_below.cells
        else:
            assert row_above is not None
            row_below_cells = [DummyCell()] * len(row_above.cells)

        cells_above = [DummyCell(), *(row_above_cells), DummyCell()]
        cells_below = [DummyCell(), *(row_below_cells), DummyCell()]
        edges = ""
        for i, ((nw, ne), (sw, se)) in enumerate(
            zip(
                pairwise(cells_above),
                pairwise(cells_below),
            )
        ):
            output += [
                (
                    self.border_style,
                    self.get_node(
                        nw.weighted_border,
                        ne_wb := ne.weighted_border,
                        se_wb := se.weighted_border,
                        sw.weighted_border,
                    ),
                )
            ]
            if i < len(col_widths):
                edge = self.get_horizontal_edge(ne_wb, se_wb)
                edges += edge
                style = self.border_style
                # Use cell style for borders in colspans
                if isinstance(se, SpacerCell) and se.expands.rowspan > 1:
                    style = se.expands.style
                output += [(style, edge * (col_widths[i]))]
        # Do not draw border row if border collapse is on and all parts are invisible
        if border_collapse and not edges.strip():
            return []
        output += [("", "\n")]
        return output

    def draw_table_row(
        self,
        row: "Optional[RowCol]",
        col_widths: "list[int]",
        border_collapse: "bool",
    ) -> "StyleAndTextTuples":
        """Draws a row in the table."""
        output: "StyleAndTextTuples" = []
        if row:
            # Calculate borders
            borders = []
            for e, w in zip([DummyCell(), *row.cells], [*row.cells, DummyCell()]):
                borders += [
                    (
                        self.border_style,
                        self.get_vertical_edge(e.weighted_border, w.weighted_border),
                    )
                ]
            if self.border_collapse and not "".join([x[1] for x in borders]).strip():
                borders = [("", "")] * len(borders)

            def _calc_cell_width(
                cell: "Cell", col: "int", col_widths: "list[int]"
            ) -> "int":
                """Calculate a cell's width."""
                if isinstance(cell, SpacerCell) and cell.expands.colspan > 1:
                    return 0
                width = col_widths[col]
                # Remove padding
                padding = cell.padding
                width -= (padding.left or 0) + (padding.right or 0)

                # Expand if colspan
                for i in range(col + 1, col + cell.colspan):
                    width += col_widths[i]
                    if not border_collapse:
                        if borders[col]:
                            width += 1

                return width

            # Draw row contents line by line
            row_lines = zip_longest(
                *(
                    cell.lines(width=_calc_cell_width(cell, col, col_widths))
                    for col, cell in enumerate(row.cells)
                )
            )

            for row_line in row_lines:
                for i, (line, cell) in enumerate(zip(row_line, row.cells)):
                    # Skip spacer cells
                    if isinstance(cell, SpacerCell) and cell.expands.colspan > 1:
                        continue
                    output += [borders[i]]
                    if line is not None:
                        padding_style = f"{cell.style} nounderline"
                        padding = cell.padding
                        output += [
                            (padding_style, " " * (padding.left or 0)),
                            *line,
                            (padding_style, " " * (padding.right or 0)),
                        ]
                    # Ensure last line of a cell is fill with spaces
                    else:
                        width = sum(col_widths[i : i + cell.colspan])
                        if line:
                            width -= fragment_list_width(line)
                        output.append((cell.style, " " * width))

                output += [borders[i + 1]]
                output += [("", "\n")]

        return output

    def render(self, width: "Optional[int]" = None) -> "StyleAndTextTuples":
        """Draws the table, optionally at a given character width."""
        width = width or self.width
        col_widths = self.calculate_col_widths(width)
        output = []

        if self.rows:
            border_collapse = self.border_collapse
            for row_above, row_below in zip([None, *self.rows], [*self.rows, None]):
                output += self.draw_border_row(
                    row_above, row_below, col_widths, border_collapse
                )
                output += self.draw_table_row(row_below, col_widths, border_collapse)
            if output:
                output.pop()

        return output

    def __pt_formatted_text__(self) -> "StyleAndTextTuples":
        """Render the table as formatted text."""
        return self.render()


class DummyRow(Row):
    """A dummy row - created to hold cells without an assigned column."""

    def add_cell(self, cell: "Cell") -> "None":
        """Prevents cells being added to a dummy row."""
        raise NotImplementedError("Cannot add a cell to a DummyRow")


class DummyCol(Col):
    """A dummy column - created to hold cells without an assigned column."""

    def add_cell(self, cell: "Cell") -> "None":
        """Prevents cells being added to a dummy column."""
        raise NotImplementedError("Cannot add a cell to a DummyCol")


class DummyTable(Table):
    """A dummy table - created to hold rows and columns without an assigned table."""

    def __init__(self, *args: "Any", **kwargs: "Any"):
        """Create a new dummy table."""
        kwargs["border"] = Invisible
        super().__init__(*args, **kwargs)

    def add_row(self, row: "Row") -> "None":
        """Prevents rows being added to a dummy table."""
        raise NotImplementedError("Cannot add a row to a DummyTable")

    def add_col(self, col: "Col") -> "None":
        """Prevents columns being added to a dummy table."""
        raise NotImplementedError("Cannot add a column to a DummyTable")
